#-*- encoding = utf-8 -*-
# android keycode

KEYCODE_UNKNOWN = 0
KEYCODE_SOFT_LEFT = 1
KEYCODE_SOFT_RIGHT = 2
KEYCODE_HOME = 3
KEYCODE_BACK = 4
KEYCODE_CALL = 5
KEYCODE_ENDCALL = 6
KEYCODE_0 = 7
KEYCODE_1 = 8
KEYCODE_2 = 9
KEYCODE_3 = 10
KEYCODE_4 = 11
KEYCODE_5 = 12
KEYCODE_6 = 13
KEYCODE_7 = 14
KEYCODE_8 = 15
KEYCODE_9 = 16
KEYCODE_STAR = 17
KEYCODE_POUND = 18
KEYCODE_DPAD_UP = 19
KEYCODE_DPAD_DOWN = 20
KEYCODE_DPAD_LEFT = 21
KEYCODE_DPAD_RIGHT = 22
KEYCODE_DPAD_CENTER = 23
KEYCODE_VOLUME_UP = 24
KEYCODE_VOLUME_DOWN = 25
KEYCODE_POWER = 26
KEYCODE_CAMERA = 27
KEYCODE_CLEAR = 28
KEYCODE_A = 29
KEYCODE_B = 30
KEYCODE_C = 31
KEYCODE_D = 32
KEYCODE_E = 33
KEYCODE_F = 34
KEYCODE_G = 35
KEYCODE_H = 36
KEYCODE_I = 37
KEYCODE_J = 38
KEYCODE_K = 39
KEYCODE_L = 40
KEYCODE_M = 41
KEYCODE_N = 42
KEYCODE_O = 43
KEYCODE_P = 44
KEYCODE_Q = 45
KEYCODE_R = 46
KEYCODE_S = 47
KEYCODE_T = 48
KEYCODE_U = 49
KEYCODE_V = 50
KEYCODE_W = 51
KEYCODE_X = 52
KEYCODE_Y = 53
KEYCODE_Z = 54
KEYCODE_COMMA = 55
KEYCODE_PERIOD = 56
KEYCODE_ALT_LEFT = 57
KEYCODE_ALT_RIGHT = 58
KEYCODE_SHIFT_LEFT = 59
KEYCODE_SHIFT_RIGHT = 60
KEYCODE_TAB = 61
KEYCODE_SPACE = 62
KEYCODE_SYM = 63
KEYCODE_EXPLORER = 64
KEYCODE_ENVELOPE = 65
KEYCODE_ENTER = 66
KEYCODE_DEL = 67
KEYCODE_GRAVE = 68
KEYCODE_MINUS = 69
KEYCODE_EQUALS = 70
KEYCODE_LEFT_BRACKET = 71
KEYCODE_RIGHT_BRACKET = 72
KEYCODE_BACKSLASH = 73
KEYCODE_SEMICOLON = 74
KEYCODE_APOSTROPHE = 75
KEYCODE_SLASH = 76
KEYCODE_AT = 77
KEYCODE_NUM = 78
KEYCODE_HEADSETHOOK = 79
KEYCODE_FOCUS = 80
KEYCODE_PLUS = 81
KEYCODE_MENU = 82
KEYCODE_NOTIFICATION = 83
KEYCODE_SEARCH = 84
KEYCODE_MEDIA_PLAY_PAUSE = 85
KEYCODE_MEDIA_STOP = 86
KEYCODE_MEDIA_NEXT = 87
KEYCODE_MEDIA_PREVIOUS = 88
KEYCODE_MEDIA_REWIND = 89
KEYCODE_MEDIA_FAST_FORWARD = 90
KEYCODE_MUTE = 91
KEYCODE_PAGE_UP = 92
KEYCODE_PAGE_DOWN = 93
KEYCODE_PICTSYMBOLS = 94
KEYCODE_SWITCH_CHARSET = 95
KEYCODE_BUTTON_A = 96
KEYCODE_BUTTON_B = 97
KEYCODE_BUTTON_C = 98
KEYCODE_BUTTON_X = 99
KEYCODE_BUTTON_Y = 100
KEYCODE_BUTTON_Z = 101
KEYCODE_BUTTON_L1 = 102
KEYCODE_BUTTON_R1 = 103
KEYCODE_BUTTON_L2 = 104
KEYCODE_BUTTON_R2 = 105
KEYCODE_BUTTON_THUMBL = 106
KEYCODE_BUTTON_THUMBR = 107
KEYCODE_BUTTON_START = 108
KEYCODE_BUTTON_SELECT = 109
KEYCODE_BUTTON_MODE = 110
KEYCODE_ESCAPE = 111
KEYCODE_FORWARD_DEL = 112
KEYCODE_CTRL_LEFT = 113
KEYCODE_CTRL_RIGHT = 114
KEYCODE_CAPS_LOCK = 115
KEYCODE_SCROLL_LOCK = 116
KEYCODE_META_LEFT = 117
KEYCODE_META_RIGHT = 118
KEYCODE_FUNCTION = 119
KEYCODE_SYSRQ = 120
KEYCODE_BREAK = 121
KEYCODE_MOVE_HOME = 122
KEYCODE_MOVE_END = 123
KEYCODE_INSERT = 124
KEYCODE_FORWARD = 125
KEYCODE_MEDIA_PLAY = 126
KEYCODE_MEDIA_PAUSE = 127
KEYCODE_MEDIA_CLOSE = 128
KEYCODE_MEDIA_EJECT = 129
KEYCODE_MEDIA_RECORD = 130
KEYCODE_F1 = 131
KEYCODE_F2 = 132
KEYCODE_F3 = 133
KEYCODE_F4 = 134
KEYCODE_F5 = 135
KEYCODE_F6 = 136
KEYCODE_F7 = 137
KEYCODE_F8 = 138
KEYCODE_F9 = 139
KEYCODE_F10 = 140
KEYCODE_F11 = 141
KEYCODE_F12 = 142
KEYCODE_NUM_LOCK = 143
KEYCODE_NUMPAD_0 = 144
KEYCODE_NUMPAD_1 = 145
KEYCODE_NUMPAD_2 = 146
KEYCODE_NUMPAD_3 = 147
KEYCODE_NUMPAD_4 = 148
KEYCODE_NUMPAD_5 = 149
KEYCODE_NUMPAD_6 = 150
KEYCODE_NUMPAD_7 = 151
KEYCODE_NUMPAD_8 = 152
KEYCODE_NUMPAD_9 = 153
KEYCODE_NUMPAD_DIVIDE = 154
KEYCODE_NUMPAD_MULTIPLY = 155
KEYCODE_NUMPAD_SUBTRACT = 156
KEYCODE_NUMPAD_ADD = 157
KEYCODE_NUMPAD_DOT = 158
KEYCODE_NUMPAD_COMMA = 159
KEYCODE_NUMPAD_ENTER = 160
KEYCODE_NUMPAD_EQUALS = 161
KEYCODE_NUMPAD_LEFT_PAREN = 162
KEYCODE_NUMPAD_RIGHT_PAREN = 163
KEYCODE_VOLUME_MUTE = 164
KEYCODE_INFO = 165
KEYCODE_CHANNEL_UP = 166
KEYCODE_CHANNEL_DOWN = 167
KEYCODE_ZOOM_IN = 168
KEYCODE_ZOOM_OUT = 169
KEYCODE_TV = 170
KEYCODE_WINDOW = 171
KEYCODE_GUIDE = 172
KEYCODE_DVR = 173
KEYCODE_BOOKMARK = 174
KEYCODE_CAPTIONS = 175
KEYCODE_SETTINGS = 176
KEYCODE_TV_POWER = 177
KEYCODE_TV_INPUT = 178
KEYCODE_STB_POWER = 179
KEYCODE_STB_INPUT = 180
KEYCODE_AVR_POWER = 181
KEYCODE_AVR_INPUT = 182
KEYCODE_PROG_RED = 183
KEYCODE_PROG_GREEN = 184
KEYCODE_PROG_YELLOW = 185
KEYCODE_PROG_BLUE = 186
KEYCODE_APP_SWITCH = 187
KEYCODE_BUTTON_1 = 188
KEYCODE_BUTTON_2 = 189
KEYCODE_BUTTON_3 = 190
KEYCODE_BUTTON_4 = 191
KEYCODE_BUTTON_5 = 192
KEYCODE_BUTTON_6 = 193
KEYCODE_BUTTON_7 = 194
KEYCODE_BUTTON_8 = 195
KEYCODE_BUTTON_9 = 196
KEYCODE_BUTTON_10 = 197
KEYCODE_BUTTON_11 = 198
KEYCODE_BUTTON_12 = 199
KEYCODE_BUTTON_13 = 200
KEYCODE_BUTTON_14 = 201
KEYCODE_BUTTON_15 = 202
KEYCODE_BUTTON_16 = 203
KEYCODE_LANGUAGE_SWITCH = 204
KEYCODE_MANNER_MODE = 205
KEYCODE_3D_MODE = 206
KEYCODE_CONTACTS = 207
KEYCODE_CALENDAR = 208
KEYCODE_MUSIC = 209
KEYCODE_CALCULATOR = 210
KEYCODE_ZENKAKU_HANKAKU = 211
KEYCODE_EISU = 212
KEYCODE_MUHENKAN = 213
KEYCODE_HENKAN = 214
KEYCODE_KATAKANA_HIRAGANA = 215
KEYCODE_YEN = 216
KEYCODE_RO = 217
KEYCODE_KANA = 218
KEYCODE_ASSIST = 219
KEYCODE_BRIGHTNESS_DOWN = 220
KEYCODE_BRIGHTNESS_UP = 221
KEYCODE_MEDIA_AUDIO_TRACK = 222

KEYBOARD_KEYS = {
    '\x1b'  : KEYCODE_ESCAPE,    
    '\r'    : KEYCODE_ENTER,
    '\x7f'  : KEYCODE_DEL,
    '\t'    : KEYCODE_TAB,
    ' '     : KEYCODE_SPACE,
    '-'     : KEYCODE_MINUS,
    '='     : KEYCODE_EQUALS,
    '['     : KEYCODE_LEFT_BRACKET,
    ']'     : KEYCODE_RIGHT_BRACKET,
    '\\'    : KEYCODE_BACKSLASH,
    ';'     : KEYCODE_SEMICOLON,
    "'"     : KEYCODE_APOSTROPHE,
    '`'     : KEYCODE_GRAVE,
    ','     : KEYCODE_COMMA,
    '.'     : KEYCODE_PERIOD,
    '/'     : KEYCODE_SLASH,
}

import sys
if sys.platform == 'win32':
    KEYBOARD_KEYS['\x08'] = KEYCODE_DEL # Oh No!

SHIFTED_KEYS = {
    '!' : KEYCODE_1,    
    '@' : KEYCODE_2,
    '#' : KEYCODE_3,
    '$' : KEYCODE_4,
    '%' : KEYCODE_5,
    '^' : KEYCODE_6,
    '&' : KEYCODE_7,
    '*' : KEYCODE_8,
    '(' : KEYCODE_9,
    ')' : KEYCODE_0,
    '_' : KEYCODE_MINUS,
    '+' : KEYCODE_EQUALS,
    '{' : KEYCODE_LEFT_BRACKET,
    '}' : KEYCODE_RIGHT_BRACKET,
    '|' : KEYCODE_BACKSLASH,
    ':' : KEYCODE_SEMICOLON,
    '"' : KEYCODE_APOSTROPHE,
    '~' : KEYCODE_GRAVE,
    '<' : KEYCODE_COMMA,
    '>' : KEYCODE_PERIOD,
    '?' : KEYCODE_SLASH,
}

CTRLED_KEYS = {
    '\x01' : KEYCODE_A,
    '\x02' : KEYCODE_B,
    '\x03' : KEYCODE_C,
    '\x04' : KEYCODE_D,
    '\x05' : KEYCODE_E,
    '\x06' : KEYCODE_F,
    '\x07' : KEYCODE_G,
    '\x08' : KEYCODE_H,
    '\t' : KEYCODE_I,
    '\n' : KEYCODE_J,
    '\x0b' : KEYCODE_K,
    '\x0c' : KEYCODE_L,
    '\r' : KEYCODE_M,
    '\x0e' : KEYCODE_N,
    '\x0f' : KEYCODE_O,
    '\x10' : KEYCODE_P,
    '\x11' : KEYCODE_Q,
    '\x12' : KEYCODE_R,
    '\x13' : KEYCODE_S,
    '\x14' : KEYCODE_T,
    '\x15' : KEYCODE_U,
    '\x16' : KEYCODE_V,
    '\x17' : KEYCODE_W,
    '\x18' : KEYCODE_X,
    '\x19' : KEYCODE_Y,
    '\x1a' : KEYCODE_Z,
}



# -*- coding: utf-8 -*-
# An android event hook via getevent.
# Only ABS_MT_POSITION_X(Y) events are handled.
#
# Basic input: TouchDown(D), TouchUp(U), TouchMove(M)
# Basic timeouts: TouchPressTimeout(P), TouchFollowTimeout(F), TouchMoveStopTimeout(S)
# guestures are defined as follows:
#   Tap/Touch/Click: DM?UF
#   TapFollow: (DM?U)+DM?UF
#   LongPress: DP, may be followed by Drag or Swipe
#   Drag: D?M+S, may be followed by Drag or Swipe
#   Swipe/Fling: D?M+U, difference with `Drag` is that `TouchMoveStopTimeout` cannot be fired.
#   2-Finger-Pinch: distance changing
#   2-Finger-Drag: distance hold while moving
# where '?' after M means a little movement and '+' means a large one.
# other guestures are ignored.

import re
import math
import time
import numpy as np
import subprocess
import threading
import Queue
import traceback

__all__ = ['AndroidInputHookManager', 'HookManager', 'HookConstants']

# global, max MultiTap count. Set to 1 to disable MultiTap, 0 for infinite.
_MULTI_TAP_NUM = 3

def set_multitap(count):
    if count < 0:
        print 'Cannot set to negative count.'
        return
    global _MULTI_TAP_NUM
    _MULTI_TAP_NUM = int(count)

class HookConstants:
    # basic events
    TOUCH_ANY  = 1 << 3
    TOUCH_DOWN = 1 << 3 ^ 1
    TOUCH_UP   = 1 << 3 ^ 2
    TOUCH_MOVE = 1 << 3 ^ 3
    # only used for gesture analyze
    TOUCH_PRESS_TIMEOUT     = 1 << 3 ^ 4
    TOUCH_FOLLOW_TIMEOUT    = 1 << 3 ^ 5
    TOUCH_MOVESTOP_TIMEOUT  = 1 << 3 ^ 6

    # DOWN is odd, UP is even & DONW + 1 == UP
    KEY_ANY = 1 << 4
    KEY_HOME_DOWN = 1 << 4 ^ 1
    KEY_HOME_UP   = 1 << 4 ^ 2
    KEY_BACK_DOWN = 1 << 4 ^ 3
    KEY_BACK_UP   = 1 << 4 ^ 4
    KEY_MENU_DOWN = 1 << 4 ^ 5
    KEY_MENU_UP   = 1 << 4 ^ 6
    KEY_POWER_DOWN      = 1 << 4 ^ 7
    KEY_POWER_UP        = 1 << 4 ^ 8
    KEY_VOLUMEDOWN_DOWN = 1 << 4 ^ 9
    KEY_VOLUMEDOWN_UP   = 1 << 4 ^ 10
    KEY_VOLUMEUP_DOWN   = 1 << 4 ^ 11
    KEY_VOLUMEUP_UP     = 1 << 4 ^ 12

    # gestures
    GST_TAP         = 1 << 5 ^ 1
    GST_MULTI_TAP   = 1 << 5 ^ 2
    GST_LONG_PRESS  = 1 << 5 ^ 3
    GST_LONG_PRESS_RELEASE = 1 << 5 ^ 4
    GST_DRAG        = 1 << 5 ^ 5
    GST_SWIPE       = 1 << 5 ^ 6
    GST_PINCH_IN    = 1 << 5 ^ 7
    GST_PINCH_OUT   = 1 << 5 ^ 8

HC = HookConstants

HCREPR = {
    HC.TOUCH_DOWN : 'D',
    HC.TOUCH_UP   : 'U',
    HC.TOUCH_MOVE : 'M',
    HC.TOUCH_PRESS_TIMEOUT     : 'P',
    HC.TOUCH_FOLLOW_TIMEOUT    : 'F',
    HC.TOUCH_MOVESTOP_TIMEOUT  : 'S',
    HC.GST_TAP: 'Tap',
    HC.GST_MULTI_TAP: 'MultiTap',
    HC.GST_LONG_PRESS: 'LongPress',
    HC.GST_LONG_PRESS_RELEASE: 'PressRelease',
    HC.GST_DRAG: 'Drag',
    HC.GST_SWIPE: 'Swipe',
    HC.GST_PINCH_IN: 'PinchIn',
    HC.GST_PINCH_OUT: 'PinchOut',
}

class Event(object):
    def __init__(self, time, msg):
        self.time = time
        self.msg = msg

    def __str__(self):
        return '%s_%s' % (self.__class__.__name__, HCREPR.get(self.msg, self.msg))

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__,
            ', '.join(['%s=%s' % (k, v) for k, v in self.__dict__.iteritems()]))

class KeyEvent(Event):
    def __init__(self, time, msg, key):
        super(KeyEvent, self).__init__(time, msg)
        # convert to KEYCODE_xxx for 'adb input keyevent xxx'
        if key.startswith('KEY_'):
            key = 'KEYCODE_' + key[4:]
        self.key = key

class TouchEvent(Event):
    def __init__(self, time, msg, slotid, x, y, pressure, touch_major, **extra):
        super(TouchEvent, self).__init__(time, msg)
        self.slotid = slotid
        self.x = x
        self.y = y
        self.pressure = pressure
        self.touch_major = touch_major
        self.__dict__.update(extra)

class TouchTimeoutEvent(Event):
    def __init__(self, time, msg, slotid):
        super(TouchTimeoutEvent, self).__init__(time, msg)
        self.slotid = slotid

class GestureEvent(Event):
    def __init__(self, msg, track):
        # suffixes: s for start, e for end.
        # two-finger guestures need two tracks
        if msg in (HC.GST_PINCH_IN, HC.GST_PINCH_OUT):
            t1, t2 = track[0], track[1]
            ts = min(t1[0].time, t2[0].time)
            te = max(t1[-1].time, t2[-1].time)
        else:
            es, ee = track[0], track[-1]
            ts, te = track[0].time, track[-1].time
            print 'Gesture', HCREPR.get(msg, msg), ''.join([HCREPR.get(e.msg, e.msg) for e in track]), (es.x, es.y), (ee.x, ee.y)
            if msg in (HC.GST_SWIPE, HC.GST_DRAG):
                # TODO: check for corners for complicated trace
                self.points = [(es.x, es.y), (ee.x, ee.y)]
            else:
                self.points = [(es.x, es.y), (ee.x, ee.y)]

        super(GestureEvent, self).__init__(ts, msg)
        self.duration = te - ts

SLOT_NUM = 5
_X, _Y, _VR, _VA, _MJ, _PR, FIELD_NUM = range(7)
INF = 9999

class InputParser(object):
    _pat = re.compile('\[\s*(?P<time>[0-9.]+)\] (?P<device>/dev/.*): +(?P<type>\w+) +(?P<code>\w+) +(?P<value>\w+)')
    _move_radius = 10

    def __init__(self, queue):
        self.timediff = None
        self.queue = queue
        # the 'standard' status temp_status is compared to.
        # if changes are great enough, new event are emitted.
        # velocity will be calculated for touch-move events.
        self._status = np.ones((SLOT_NUM, FIELD_NUM), dtype=int) * (-INF)
        self._status_time = 0
        # realtime status, minor changes are cumulated
        self._temp_status = np.ones((SLOT_NUM, FIELD_NUM), dtype=int) * (-INF)
        self._temp_status_time = 0

        self._touch_batch = []
        self._curr_slot = 0

    def feed(self, line):
        # print line
        m = self._pat.search(line)
        if not m:
            return
        _time, _device, _type, _code, _value = m.groups()
        _time = float(_time)
        if self.timediff is None:
            self.timediff = time.time() - _time
        _time = self.timediff + _time
        try:
            _value = int(_value, 16)
        except:
            pass

        if _type == 'EV_SYN':
            if _code in ('SYN_REPORT', 'SYN_MT_REPORT'):
                try:
                    self._process_touch_batch()
                except IndexError: # there might be a 6th finger, ignore that.
                    self._touch_batch = []
            elif _code == 'SYN_DROPPED':
                self._touch_batch = []
            else:
                # print 'unknown syn code', _code
                pass
        elif _type == 'EV_KEY':
            self.emit_key_event(_time, _code, _value)
        elif _type == 'EV_ABS':
            self._touch_batch.append((_time, _device, _type, _code, _value))
        else:
            # print 'unknown input event type', _type
            pass

    def emit_key_event(self, _time, _code, _value):
        name = '%s_%s' % (_code, _value)
        msg = getattr(HC, name, None)
        if msg is None:
            return
        event = KeyEvent(_time, msg, _code)
        self.queue.put(event)

    def emit_touch_event(self, event):
        self.queue.put(event)

    def _process_touch_batch(self):
        '''a batch syncs in about 0.001 seconds.'''
        if not self._touch_batch:
            return

        _time = self._temp_status_time
        changed = False

        for (_time, _device, _type, _code, _value) in self._touch_batch:
            if _code == 'ABS_MT_TRACKING_ID':
                if _value == 0xffffffff:
                    self._temp_status[self._curr_slot] = -INF
                    changed = True
                else:
                    pass
            elif _code == 'ABS_MT_SLOT':
                self._curr_slot = _value
            else:
                if _code == 'ABS_MT_POSITION_X':
                    self._temp_status[self._curr_slot,_X] = _value
                    changed = True
                elif _code == 'ABS_MT_POSITION_Y':
                    self._temp_status[self._curr_slot,_Y] = _value
                    changed = True
                elif _code == 'ABS_MT_PRESSURE':
                    self._temp_status[self._curr_slot,_PR] = _value
                elif _code == 'ABS_MT_TOUCH_MAJOR':
                    self._temp_status[self._curr_slot,_MJ] = _value
                else:
                    print 'Unknown code', _code

        self._temp_status_time = _time
        self._touch_batch = []
        if not changed:
            return

        # check differences, if position changes are big enough then emit events
        diff = self._temp_status - self._status
        dt = self._temp_status_time - self._status_time
        emitted = False
        for i in range(SLOT_NUM):
            arr = self._temp_status[i]
            oldarr = self._status[i]
            dx, dy = diff[i,_X], diff[i,_Y]
            if dx > INF or dy > INF:
                # touch begin
                event = TouchEvent(_time, HC.TOUCH_DOWN, i, arr[_X], arr[_Y], arr[_PR], arr[_MJ])
                self.emit_touch_event(event)
                emitted = True
            elif dx < -INF or dy < -INF:
                # touch end
                event = TouchEvent(_time, HC.TOUCH_UP, i, oldarr[_X], oldarr[_Y], oldarr[_PR], oldarr[_MJ])
                self.emit_touch_event(event)
                emitted = True
            else:
                r, a = radang(float(dx), float(dy))
                if r > self._move_radius:
                    v = r / dt
                    event = TouchEvent(_time, HC.TOUCH_MOVE, i, arr[_X], arr[_Y], arr[_PR], arr[_MJ], angle=a, velocity=v)
                    self.emit_touch_event(event)
                    emitted = True

        if not emitted:
            return
        self._status = self._temp_status.copy()
        self._status_time = self._temp_status_time

def radang(x, y):
    '''return (radius, angle) of a vector(x, y)'''
    if x == 0:
        if y == 0:
            return 0, 0
        return abs(y), 90+180*(y<0)
    if y == 0:
        return abs(x), 180*(x<0)

    r = math.sqrt(x*x+y*y)
    a = math.degrees(math.atan(y/x))
    if x < 0:
        a += 180
    elif y < 0:
        a += 360
    return r, a

class GestureRecognizer(object):

    double_tap_delay = 0.5
    long_press_delay = 1
    move_stop_delay = 0.2
    pinch_difference_square = 3000

    def __init__(self, queue):
        self.queue = queue
        self.dispatch_map = {}
        self.running = False

        self.touches = [None] * SLOT_NUM

        # used for recognition
        self.tracks = [None for i in range(SLOT_NUM)]
        self.track_slots = set()

    def register(self, keycode, func):
        self.dispatch_map[keycode] = func

    def start(self):
        if self.running:
            return
        self.running = True
        t = threading.Thread(target=self.process)
        t.setDaemon(True)
        t.start()

    def stop(self):
        self.running = False

    def process(self):
        '''handle events and trigger time-related events'''
        timediff = 0
        while True:
            try:
                time.sleep(0.001)
                event = self.queue.get_nowait()
                self.handle_event(event)
                if event.msg & HC.KEY_ANY:
                    continue
                if timediff == 0:
                    timediff = time.time() - event.time
                self.touches[event.slotid] = event
            except Queue.Empty:
                if not self.running:
                    break
                now = time.time() - timediff
                for i in range(SLOT_NUM):
                    e = self.touches[i]
                    if e is None:
                        continue
                    if e.msg == HC.TOUCH_DOWN and now - e.time > self.long_press_delay:
                        self.analyze_tracks(TouchTimeoutEvent(now, HC.TOUCH_PRESS_TIMEOUT, i))
                        self.touches[i] = None
                    elif e.msg == HC.TOUCH_UP and now - e.time > self.double_tap_delay:
                        self.analyze_tracks(TouchTimeoutEvent(now, HC.TOUCH_FOLLOW_TIMEOUT, i))
                        self.touches[i] = None
                    elif e.msg == HC.TOUCH_MOVE and now - e.time > self.move_stop_delay:
                        self.analyze_tracks(TouchTimeoutEvent(now, HC.TOUCH_MOVESTOP_TIMEOUT, i))
                        self.touches[i] = None
            except:
                traceback.print_exc()

        print 'process done.'

    def handle_event(self, event):
        self.dispatch_event(event.msg, event)
        if event.msg & HC.KEY_ANY:
            self.dispatch_event(HC.KEY_ANY, event)
        else:
            self.dispatch_event(HC.TOUCH_ANY, event)
            self.analyze_tracks(event)

    def dispatch_event(self, msg, event):
        func = self.dispatch_map.get(msg)
        if func is None:
            return
        try:
            func(event)
        except:
            traceback.print_exc()

    def analyze_tracks(self, event):
        pass

    def handle_gesture(self, msg, tracks):
        event = GestureEvent(msg, tracks)
        func = self.dispatch_map.get(msg)
        if func is None:
            return
        try:
            func(event)
        except:
            traceback.print_exc()

## NOT COMPLEMENTED ##
class SimpleGestureRecognizer(GestureRecognizer):

    N_FINGER = 2

    def analyze_tracks(self, event):
        # handle one-finger and two-finger gestures only
        # means a third finger will be ignored even if one of the
        # first two fingers leaves the screen.
        i = event.slotid

        # begin guesture when touch down
        if event.msg == HC.TOUCH_DOWN:
            if len(self.track_slots) == self.N_FINGER and i not in self.track_slots:
                return
            if self.tracks[i] is None:
                self.tracks[i] = []
                self.track_slots.add(i)
            self.tracks[i].append(event)
            return

        if self.tracks[i] is None:
            return

        if event.msg == HC.TOUCH_FOLLOW_TIMEOUT:
            self.tracks[i] = []
        elif event.msg == HC.TOUCH_PRESS_TIMEOUT:
            # print ''.join([HCREPR.get(e.msg) for e in self.tracks[i]]), 'long press'
            self.tracks[i] = []
        elif event.msg == HC.TOUCH_MOVESTOP_TIMEOUT:
            # print ''.join([HCREPR.get(e.msg) for e in self.tracks[i]]), 'drag'
            self.tracks[i] = []
            if len(self.track_slots) == 2:
                for s in self.track_slots:
                    print s, ''.join([HCREPR.get(e.msg) for e in self.tracks[s]])
                print
        elif event.msg == HC.TOUCH_UP:
            self.tracks[i].append(event)
            if len(self.track_slots) == 2:
                for s in self.track_slots:
                    print s, ''.join([HCREPR.get(e.msg) for e in self.tracks[s]])
                print
            self.tracks[i] = None
            self.track_slots.discard(i)
        else: # TOUCH_MOVE
            self.tracks[i].append(event)
            return

            # check for pinch/pan
            if len(self.track_slots) == 2:
                t1, t2 = [self.tracks[s] for s in self.track_slots]
                if len(t1) == 0 or len(t2) == 0 or len(t1) + len(t2) < 6:
                    return
                # make copy and check distance changing
                t1, t2 = t1[:], t2[:]
                dists = []
                while len(dists) < 5:
                    e1, e2 = t1[-1], t2[-1]
                    dx, dy = e1.x-e2.x, e1.y-e2.y
                    dists.append(dx*dx+dy*dy)
                    if e1.time < e2.time:
                        if len(t2) == 1:
                            break
                        else:
                            t2.pop()
                    else:
                        if len(t1) == 1:
                            break
                        else:
                            t1.pop()
                print [dists[j+1]-dists[j] for j in range(len(dists)-1)]
                # just keep latest position
                for s in self.track_slots:
                    self.tracks[s] = self.tracks[s][-1:]

class RegexpGestureRecognizer(GestureRecognizer):

    N_FINGER = 1

    def analyze_tracks(self, event):
        # handle one-finger gestures only
        i = event.slotid

        # begin guesture when touch down
        if event.msg == HC.TOUCH_DOWN:
            if len(self.track_slots) == self.N_FINGER and i not in self.track_slots:
                return
            if not self.tracks[i]:
                self.tracks[i] = []
                self.track_slots.add(i)
            self.tracks[i].append(event)
            return

        if self.tracks[i] is None:
            return

        s = ''.join([HCREPR.get(e.msg) for e in self.tracks[i]])

        if event.msg == HC.TOUCH_FOLLOW_TIMEOUT:
            if re.match('^DM?U$', s):
                self.handle_gesture(HC.GST_TAP, self.tracks[i][:])
            elif re.match('^(DM?U)+DM?U$', s):
                self.handle_gesture(HC.GST_MULTI_TAP, self.tracks[i][:])
            self.tracks[i] = None
            self.track_slots.discard(i)
        elif event.msg == HC.TOUCH_MOVESTOP_TIMEOUT:
            if re.match('^D?MM+$', s):
                self.handle_gesture(HC.GST_DRAG, self.tracks[i][:])
            self.tracks[i] = []
        elif event.msg == HC.TOUCH_PRESS_TIMEOUT:
            if s == 'D':
                self.handle_gesture(HC.GST_LONG_PRESS, self.tracks[i][:])
            self.tracks[i] = []
        elif event.msg == HC.TOUCH_UP:
            self.tracks[i].append(event) # note: it's not the same with s after add
            if s == '':
                self.handle_gesture(HC.GST_LONG_PRESS_RELEASE, [event])
            elif re.match('^D?MM+$', s):
                self.handle_gesture(HC.GST_SWIPE, self.tracks[i][:])
                self.tracks[i] = []
            elif _MULTI_TAP_NUM == 1 and re.match('^DM?$', s):
                self.handle_gesture(HC.GST_TAP, self.tracks[i][:])
                self.tracks[i] = []
            elif _MULTI_TAP_NUM > 1 and re.match('^(DM?U){%d}DM?$' % (_MULTI_TAP_NUM-1,), s):
                self.handle_gesture(HC.GST_MULTI_TAP, self.tracks[i][:])
                self.tracks[i] = []
        elif event.msg == HC.TOUCH_MOVE:
            if re.match('^(DU)+D$', s):
                if s == 'DUD':
                    self.handle_gesture(HC.GST_TAP, self.tracks[i][:-1])
                else:
                    self.handle_gesture(HC.GST_MULTI_TAP, self.tracks[i][:-1])
                self.tracks[i] = self.tracks[i][-1:]
            self.tracks[i].append(event)

NOTACTIVE, ACTIVE, STAGE_1, STAGE_2, TRIGGERED = range(5)

## NOT COMPLEMENTED ##
class StateMachineGestureRecognizer(GestureRecognizer):

    state_map = {
        HC.GST_TAP: {
            NOTACTIVE: { HC.TOUCH_DOWN : ACTIVE },
            ACTIVE: {
                HC.TOUCH_MOVE: STAGE_1,
                HC.TOUCH_PRESS_TIMEOUT : NOTACTIVE,
                HC.TOUCH_FOLLOW_TIMEOUT : TRIGGERED,
            },
            STAGE_1: {
                HC.TOUCH_MOVE: NOTACTIVE,
                HC.TOUCH_PRESS_TIMEOUT : NOTACTIVE,
                HC.TOUCH_FOLLOW_TIMEOUT : TRIGGERED,
            }
        },
        HC.GST_SWIPE: {
            NOTACTIVE: { HC.TOUCH_DOWN: ACTIVE },
            ACTIVE: { HC.TOUCH_UP: NOTACTIVE, HC.TOUCH_MOVE: STAGE_1},
            STAGE_1: { HC.TOUCH_UP: NOTACTIVE, HC.TOUCH_MOVE: STAGE_2 },
            STAGE_2: { HC.TOUCH_UP: TRIGGERED, HC.TOUCH_MOVESTOP_TIMEOUT: TRIGGERED},
        },
    }

    def __init__(self, queue):
        super(self.__class__, self).__init__(queue)
        self.state = {}
        for k in self.state_map:
            self.state[k] = NOTACTIVE
        print self.state_map

    def analyze_tracks(self, event):
        for k, v in self.state.iteritems():
            s = self.state_map.get(k, {}).get(v, {}).get(event.msg)
            if s is not None:
                self.state[k] = s
        triggered = False
        for k, v in self.state.iteritems():
            if v == TRIGGERED:
                print 'trigger event', k
                triggered = True
        if triggered:
            for k in self.state:
                self.state[k] = NOTACTIVE

class AndroidInputHookManager(object):

    def __init__(self, serial=None, processor_class=RegexpGestureRecognizer):
        self._serial = serial
        self.running = False
        self._queue = Queue.Queue()
        self._listener = None
        self._parser = InputParser(self._queue)
        self._processor = processor_class(self._queue)

    def set_serial(self, serial):
        self._serial = serial

    def register(self, keycode, func):
        '''register hook function'''
        self._processor.register(keycode, func)

    def hook(self):
        self._processor.start()
        self.running = True
        t = threading.Thread(target=self._run_hook)
        t.setDaemon(True)
        t.start()

    def _run_hook(self):
        cmd = ['adb']
        if self._serial:
            cmd.extend(['-s', self._serial])
        cmd.extend(['shell', 'getevent', '-lt'])

        while True:
            # start listener
            self._listener = p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
            while True:
                try:
                    line = p.stdout.readline().strip()
                    if not line:
                        if p.poll() is not None:
                            break
                        continue
                    self._parser.feed(line)
                except KeyboardInterrupt:
                    p.kill()
                except:
                    p.kill()
                    traceback.print_exc()

            if not self.running:
                break
            state = subprocess.check_output(['adb', '-s', self._serial, 'get-state']).strip()
            if state != 'device':
                print 'adb status(%s) wrong! stop hook.' % (state,)
                break
            print 'adb getevent died, reconnecting...'
            time.sleep(1)

    def unhook(self):
        self.running = False
        self._processor.stop()
        if self._listener:
            self._listener.kill()

HookManager = AndroidInputHookManager

if __name__ == '__main__':
    hm = AndroidInputHookManager(processor_class=RegexpGestureRecognizer)
    hm.hook()
    while True:
        try:
            time.sleep(0.1)
        except KeyboardInterrupt:
            break
    hm.unhook()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import absolute_import

import os
import time
import warnings

import wda
import subprocess32 as subprocess
from PIL import Image
from StringIO import StringIO

from atx.drivers.mixin import DeviceMixin, hook_wrap
from atx.drivers import Display
from atx import consts
from atx import ioskit

__dir__ = os.path.dirname(os.path.abspath(__file__))


class IOSDevice(DeviceMixin):
    def __init__(self, device_url, bundle_id=None):
        DeviceMixin.__init__(self)
        self.__device_url = device_url
        self.__scale = None
        
        self._wda = wda.Client(device_url)
        self._session = None
        self._bundle_id = None

        if bundle_id:
            self.start_app(bundle_id)
            
        # ioskit.Device.__init__(self, udid)

        # # xcodebuild -project  -scheme WebDriverAgentRunner -destination "id=1002c0174e481a651d71e3d9a89bd6f90d253446" test
        # # Test Case '-[UITestingUITests testRunner]' started.
        # xproj_dir = os.getenv('WEBDRIVERAGENT_DIR')
        # if not xproj_dir:
        #     raise RuntimeError("env-var WEBDRIVERAGENT_DIR need to be set")

        # proc = self._xcproc = subprocess.Popen(['/usr/bin/xcodebuild',
        #     '-project', 'WebDriverAgent.xcodeproj',
        #     '-scheme', 'WebDriverAgentRunner',
        #     '-destination', 'id='+self.udid, 'test'],
        #     stdout=subprocess.PIPE, stderr=subprocess.STDOUT, cwd=xproj_dir, bufsize=1, universal_newlines=True)
        # for line in iter(proc.stdout.readline, b""):
        #     print 'STDOUT:', line.strip()
        #     if 'TEST FAILED' in line:
        #         raise RuntimeError("webdriver start test failed, maybe need to unlock the keychain, try\n" + 
        #             '$ security unlock-keychain ~/Library/Keychains/login.keychain')
        #     elif "Successfully wrote Manifest cache" in line:
        #         print 'GOOD ^_^, wait 5s'
        #         time.sleep(5.0)
        #         break

    def start_app(self, bundle_id):
        """Start an application
        Args:
            - bundle_id: (string) apk bundle ID

        Returns:
            WDA session object
        """
        # if self._session is not None:
        #     self.stop_app()
        self._bundle_id = bundle_id
        self._session = self._wda.session(bundle_id)
        return self._session

    @property
    def session(self):
        if self._session is None:
            self._session = self._wda.session()
        return self._session

    def stop_app(self, *args):
        if self._session is None:
            return
        self._session.close()
        self._session = None
        self._bundle_id = None

    def __call__(self, *args, **kwargs):
        return self.session(*args, **kwargs)

    def status(self):
        """ Check if connection is ok """
        return self._wda.status()

    @property
    def display(self):
        """ Get screen width and height """
        w, h = self.session.window_size()
        return Display(w*self.scale, h*self.scale)

    @property
    def bundle_id(self):
        return self._bundle_id

    @property
    def scale(self):
        if self.__scale:
            return self.__scale
        wsize = self.session.window_size()
        # duplicate operation here. But do not want fix it now.
        self.__scale = min(self.screenshot().size)/min(wsize)
        # self.__scale = min(self.__screensize) / min(wsize)
        return self.__scale
    
    @property
    def rotation(self):
        """Rotation of device
        Returns:
            int (0-3)
        """
        rs = dict(PORTRAIT=0, LANDSCAPE=1, UIA_DEVICE_ORIENTATION_LANDSCAPERIGHT=3)
        return rs.get(self.session.orientation, 0)

    def type(self, text):
        """Type text
        Args:
            text(string): input text
        """
        self.session.send_keys(text)
        
    def click(self, x, y):
        """Simulate click operation
        Args:
            x, y(int): position
        """
        rx, ry = x/self.scale, y/self.scale
        self.session.tap(rx, ry)

    def swipe(self, x1, y1, x2, y2, duration=0.5):
        """Simulate swipe operation
        Args:
            x1, y1(int): from position
            x2, y2(int): to position
            duration(float): swipe duration, unit seconds
        """
        scale = self.scale
        x1, y1, x2, y2 = x1/scale, y1/scale, x2/scale, y2/scale
        self.session.swipe(x1, y1, x2, y2, duration)

    def home(self):
        """ Return to homescreen """
        return self._wda.home()

    def _take_screenshot(self):
        """Take a screenshot, also called by Mixin
        Args:
            - filename(string): file name to save

        Returns:
            PIL Image object
        """
        raw_png = self._wda.screenshot()
        img = Image.open(StringIO(raw_png))
        return img

    def dump_view(self):
        """Dump page XML, Note, this is a test method"""
        warnings.warn("deprecated, use source() instead", DeprecationWarning)
        return self._wda.source()

    def source(self):
        """
        Dump page XML
        """
        return self._wda.source()



# -*- coding: utf-8 -*-
# minin classes. 
# use extra resource to enhance original adb device.
# interface class should have raw_cmd method with a signature 
# as follows:
#
#   def raw_cmd(self, *args, **kwargs):
#       ...
#       return subprocess.Popen(...)
#
# where `args` is adb command arguments and kwargs are 
# subprocess keyword arguments.

import os
import Queue
import re
import socket
import struct
import subprocess
import threading
import time
import traceback

__dir__ = os.path.dirname(os.path.abspath(__file__))

class RotationWatcherMixin(object):

    __rotation = 0
    __watcher_process = None

    def open_rotation_watcher(self, on_rotation_change=None):
        package_name = 'jp.co.cyberagent.stf.rotationwatcher'
        out = self.raw_cmd('shell', 'pm', 'list', 'packages', stdout=subprocess.PIPE).communicate()[0]
        if package_name not in out:
            apkpath = os.path.join(__dir__, '..', 'vendor', 'RotationWatcher.apk')
            print 'install rotationwatcher...', apkpath
            if 0 != self.raw_cmd('install', '-r', '-t', apkpath).wait():
                print 'install rotationwatcher failed.'
                return

        if self.__watcher_process is not None:
            self.__watcher_process.kill()

        out = self.raw_cmd('shell', 'pm', 'path', package_name, stdout=subprocess.PIPE).communicate()[0]
        path = out.strip().split(':')[-1]
        p = self.raw_cmd('shell', 
            'CLASSPATH="%s"' % path, 
            'app_process',
            '/system/bin',
            'jp.co.cyberagent.stf.rotationwatcher.RotationWatcher', 
            stdout=subprocess.PIPE)
        self.__watcher_process = p

        queue = Queue.Queue()

        def _pull():
            while True:
                line = p.stdout.readline().strip()
                if not line:
                    if p.poll() is not None:
                        print 'rotationwatcher stopped'
                        break
                    continue
                queue.put(line)

        t = threading.Thread(target=_pull)
        t.setDaemon(True)
        t.start()

        def listener(value):
            try:
                self.__rotation = int(value)/90
            except:
                return
            if callable(on_rotation_change):
                on_rotation_change(self.__rotation)

        def _listen():
            while True:
                try:
                    time.sleep(0.005)
                    line = queue.get_nowait()
                    listener(line)
                except Queue.Empty:
                    if p.poll() is not None:
                        break
                    continue
                except:
                    traceback.print_exc()

        t = threading.Thread(target=_listen)
        t.setDaemon(True)
        t.start()

def str2img(jpgstr, orientation=None):
    import numpy as np
    import cv2
    arr = np.fromstring(jpgstr, np.uint8)
    img = cv2.imdecode(arr, cv2.IMREAD_COLOR)
    if orientation == 1:
        return cv2.flip(cv2.transpose(img), 0) # counter-clockwise
    if orientation == 3:
        return cv2.flip(cv2.transpose(img), 1) # clockwise
    return img

class MinicapStreamMixin(object):
    __screen = None
    __minicap_process = None

    def __install_minicap(self):
        # install minicap & minitouch
        os.system('python -m atx minicap')

    def open_minicap_stream(self, port=1313):
        # ensure minicap installed
        out = self.raw_cmd('shell', 'ls', '"/data/local/tmp/minicap"', stdout=subprocess.PIPE).communicate()[0]
        if 'No such file or directory' in out:
            self.__install_minicap()

        if self.__minicap_process is not None:
            self.__minicap_process.kill()

        # if minicap is already started, kill it first.
        out = self.raw_cmd('shell', 'ps', '-C', '/data/local/tmp/minicap', stdout=subprocess.PIPE).communicate()[0]
        out = out.strip().split('\n')
        if len(out) > 1:
            idx = out[0].split().index('PID')
            pid = out[1].split()[idx]
            print 'minicap is running, killing', pid
            self.raw_cmd('shell', 'kill', '-9', pid).wait()

        # start minicap
        out = self.raw_cmd('shell', 'LD_LIBRARY_PATH=/data/local/tmp', '/data/local/tmp/minicap', '-i', 
                    stdout=subprocess.PIPE).communicate()[0]
        m = re.search('"width": (\d+).*"height": (\d+).*"rotation": (\d+)', out, re.S)
        w, h, r = map(int, m.groups())
        w, h = min(w, h), max(w, h)
        params = '{x}x{y}@{x}x{y}/{r}'.format(x=w, y=h, r=r)
        print 'starting minicap', params

        p = self.raw_cmd('shell', 
                    'LD_LIBRARY_PATH=/data/local/tmp', 
                    '/data/local/tmp/minicap', 
                    '-P %s' % params,
                    '-S',
                    stdout=subprocess.PIPE)
        self.__minicap_process = p
        time.sleep(0.5)
        # forward to tcp port 
        self.raw_cmd('forward', 'tcp:%s' % port, 'localabstract:minicap').wait()

        queue = Queue.Queue()

        # pull data from socket
        def _pull():
            # print 'start pull', p.pid, p.poll()
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            try:
                assert p.poll() is None
                s.connect(('127.0.0.1', port))
                t = s.recv(24)
                print 'minicap connected', struct.unpack('<2B5I2B', t)
                while True:
                    frame_size = struct.unpack("<I", s.recv(4))[0]
                    trunks = []
                    recvd_size = 0
                    while recvd_size < frame_size:
                        trunk_size = min(8192, frame_size-recvd_size)
                        d = s.recv(trunk_size)
                        trunks.append(d)
                        recvd_size += len(d)
                    queue.put(''.join(trunks))
            except Exception as e:
                if not isinstance(e, struct.error):
                    traceback.print_exc()
                if p.poll() is not None:
                    print 'Process died.'
                    print p.stdout.read()
                else:
                    print 'stoping minicap ...'
                    p.kill()
            finally:
                s.close()
                self.raw_cmd('forward', '--remove', 'tcp:%s' % port).wait()

        t = threading.Thread(target=_pull)
        t.setDaemon(True)
        t.start()

        out = self.raw_cmd('shell', 'getprop', 'ro.build.version.sdk', stdout=subprocess.PIPE).communicate()[0]
        sdk = int(out.strip())
        orientation = r/90

        def _listen():
            while True:
                try:
                    time.sleep(0.005)
                    frame = queue.get_nowait()
                    if sdk <= 16:
                        img = str2img(frame, orientation)
                    else:
                        img = str2img(frame)
                    self.__screen = img
                except Queue.Empty:
                    if p.poll() is not None:
                        print 'minicap died'
                        print p.stdout.read()
                        break
                    continue
                except:
                    traceback.print_exc()

        t = threading.Thread(target=_listen)
        t.setDaemon(True)
        t.start()

    def screenshot_cv2(self):
        return self.__screen

class MinitouchStreamMixin(object):
    __touch_queue = None
    __minitouch_process = None

    def __install_minitouch(self):
        # install minicap & minitouch
        os.system('python -m atx minicap')

    def open_minitouch_stream(self, port=1111):
        if self.__touch_queue is None:
            self.__touch_queue = Queue.Queue()

        # ensure minicap installed
        out = self.raw_cmd('shell', 'ls', '"/data/local/tmp/minitouch"', stdout=subprocess.PIPE).communicate()[0]
        if 'No such file or directory' in out:
            self.__install_minitouch()

        if self.__minitouch_process is not None:
            self.__minitouch_process.kill()

        out = self.raw_cmd('shell', 'ps', '-C', '/data/local/tmp/minitouch', stdout=subprocess.PIPE).communicate()[0]
        out = out.strip().split('\n')
        if len(out) > 1:
            p = None
        else:
            p = self.raw_cmd('shell', '/data/local/tmp/minitouch')
            time.sleep(1)
            if p.poll() is not None:
                print 'start minitouch failed.'
                return
        self.__minitouch_process = p                
        self.raw_cmd('forward', 'tcp:%s' % port, 'localabstract:minitouch').wait()

        def send():
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            try:
                s.connect(('127.0.0.1', port))
                while True:
                    cmd = self.__touch_queue.get() # wait here
                    if not cmd:
                        continue
                    elif cmd[-1] != '\n':
                        cmd += '\n'
                    s.send(cmd)    
            except:
                traceback.print_exc()
            finally:
                s.close()
                self.raw_cmd('forward', '--remove', 'tcp:%s' % port).wait()

        t = threading.Thread(target=send)
        t.setDaemon(True)
        t.start()

    def click(self, x, y):
        cmd = 'd 0 %d %d 30\nc\nu 0\nc\n' % (int(x), int(y))
        self.__touch_queue.put(cmd)

    def swipe(self, sx, sy, ex, ey, steps=20):
        x1, y1, x2, y2 = map(int, (x1, y1, x2, y2))
        dx = (x2-x1)/steps
        dy = (y2-y1)/steps
        send = self.touchqueue.put
        send('d 0 %d %d 30\nc\n' % (x1, y1))
        for i in range(steps-1):
            x, y = x1+(i+1)*dx, y1+(i+1)*dy
            send('m 0 %d %d 30\nc\n' % (x, y))
        send('u 0 %d %d 30\nc\nu 0\nc\n' % (x2, y2))

    def pinchin(self, x1, y1, x2, y2, steps=10):
        pass

    def pinchout(self, x1, y1, x2, y2, steps=10):
        pass

class OpenSTFServiceMixin(object):
    pass


#-------------- examples ----------------#

class DummyDevice(object):
    def raw_cmd(self, *args, **kwargs):
        cmds = ['adb'] + list(args)
        print cmds
        return subprocess.Popen(cmds, **kwargs)

# Mixins should come in front to override functions in Base
class TestDevice(MinitouchStreamMixin, MinicapStreamMixin, RotationWatcherMixin, DummyDevice):
    def __init__(self, *args, **kwargs):
        super(self.__class__, self).__init__(*args, **kwargs)
        self.open_rotation_watcher(on_rotation_change=lambda v: self.open_minicap_stream())
        self.open_minitouch_stream()

if __name__ == '__main__':
    import cv2
    dev = TestDevice()
    while True:
        img = dev.screenshot_cv2()
        if img is not None:
            cv2.imshow('screen', img)
        cv2.waitKey(10)




#!/usr/bin/env python
# coding: utf-8
#
# > How to get tkinter canvas to dynamically resize to window width?
# http://stackoverflow.com/questions/22835289/how-to-get-tkinter-canvas-to-dynamically-resize-to-window-width
#
# > Canvas tutoril
# http://www.tkdocs.com/tutorial/canvas.html
#
# > Canvas API reference
# http://effbot.org/tkinterbook/canvas.htm
#
# > Tutorial canvas tk
# http://www.tutorialspoint.com/python/tk_canvas.htm

import Tkinter as tk
import tkSimpleDialog
from PIL import Image, ImageTk

import atx


class CropIDE(object):
    def __init__(self, title='AirtestX Basic GUI', screenshot=None):
        self._root = tk.Tk()
        self._init_items()
        self._root.title(title)
        self._lastx = 0
        self._lasty = 0
        self._bounds = None # crop area
        self._center = (0, 0) # center point, used for offset
        self._size = (90, 90)
        self._moved = False # click or click and move
        self._color = 'red' # draw color
        self._tkimage = None # keep reference
        self._image = None
        self._screenshot = screenshot
        self._ratio = 0.5

    def _init_items(self):
        root = self._root
        frm_control = tk.Frame(root, bg='blue')
        frm_control.grid(column=0, row=0)
        frm_screen = tk.Frame(root, bg='#aaa')
        frm_screen.grid(column=0, row=1)

        tk.Button(frm_control, text="Refresh", command=self._redraw).grid(column=0, row=0, sticky=tk.W)
        tk.Button(frm_control, text="Save crop", command=self._save_crop).grid(column=1, row=0, sticky=tk.W)

        self.canvas = tk.Canvas(frm_screen, bg="blue", bd=0, highlightthickness=0, relief='ridge')
        self.canvas.grid(column=0, row=0, padx=10, pady=10)
        self.canvas.bind("<Button-1>", self._stroke_start)
        self.canvas.bind("<B1-Motion>", self._stroke_move)
        self.canvas.bind("<B1-ButtonRelease>", self._stroke_done)

    def _fix_bounds(self, bounds):
        bounds = [x/self._ratio for x in bounds]
        (x0, y0, x1, y1) = bounds
        if x0 > x1:
            x0, y0, x1, y1 = x1, y1, x0, y0
        # in case of out of bounds
        w, h = self._size
        x0 = max(0, x0)
        y0 = max(0, y0)
        x1 = min(w, x1)
        y1 = min(h, y1)
        return map(int, [x0, y0, x1, y1])

    def _save_crop(self):
        print self._bounds
        if self._bounds is None:
            return
        bounds = self._fix_bounds(self._bounds)
        print bounds
        save_to = tkSimpleDialog.askstring("Save cropped image", "Enter filename")
        if save_to:
            if save_to.find('.') == -1:
                save_to += '.png'
            print('Save to:', save_to)
            self._image.crop(bounds).save(save_to)
            # cv2.imwrite(save_to, image)

    def _redraw(self):
        image = self._screenshot()
        self.draw_image(image)
        self._bounds = None
        self._reset()

    def _reset(self):
        self.canvas.delete('boundsLine')
        self.canvas.delete('clickPosition')

    def _stroke_start(self, event):
        self._moved = False
        c = self.canvas
        self._lastx, self._lasty = c.canvasx(event.x), c.canvasy(event.y)
        print 'click:', self._lastx, self._lasty

    def _stroke_move(self, event):
        self._moved = True
        self._reset()
        c = self.canvas
        x, y = c.canvasx(event.x), c.canvasy(event.y)
        self._bounds = (self._lastx, self._lasty, x, y)
        self._draw_bounds(self._bounds)
        x, y = (self._lastx+x)/2, (self._lasty+y)/2
        self.tag_point(x, y)

    def _stroke_done(self, event):
        c = self.canvas
        x, y = c.canvasx(event.x), c.canvasy(event.y)
        if self._moved:
            x, y = (self._lastx+x)/2, (self._lasty+y)/2
        self._center = (x, y) # rember position
        self.tag_point(x, y)
        self.canvas.itemconfigure('boundsLine', width=2)

    def _draw_bounds(self, bounds):
        c = self.canvas
        (x0, y0, x1, y1) = self._bounds
        c.create_rectangle(x0, y0, x1, y1, outline=self._color, tags='boundsLine', width=5)#, fill="blue")
        # c.create_line((x0, y0, x1, y1), fill=self._color, width=2, tags='boundsLine', dash=(4, 4))
        # c.create_line((x0, y1, x1, y0), fill=self._color, width=2, tags='boundsLine', dash=(4, 4))

    def tag_point(self, x, y):
        # coord = 10, 50, 110, 150
        self.canvas.delete('clickPosition')
        r = max(min(self._size)/30*self._ratio, 5)
        self.canvas.create_line(x-r, y, x+r, y, width=2, fill=self._color, tags='clickPosition')
        self.canvas.create_line(x, y-r, x, y+r, width=2, fill=self._color, tags='clickPosition')
        # coord = x-r, y-r, x+r, y+r
        # self.canvas.create_oval(coord, fill='gray', stipple="gray50", tags='clickPosition')

    def draw_image(self, image):
        self._image = image
        self._size = (width, height) = image.size
        w, h = int(width*self._ratio), int(height*self._ratio)
        # print w, h
        image = image.copy()
        image.thumbnail((w, h), Image.ANTIALIAS)
        tkimage = ImageTk.PhotoImage(image)
        self._tkimage = tkimage # keep a reference
        self.canvas.config(width=w, height=h)
        self.canvas.create_image(0, 0, anchor=tk.NW, image=tkimage)

    def mainloop(self):
        self._root.mainloop()
        

def atx_ide(serial):
    d = atx.connect(serial)
    gui = CropIDE('AirtestX IDE SN: %s' % serial, screenshot=d.screenshot)
    gui.draw_image(d.screenshot())
    gui.mainloop()

def test():
    # image = Image.open('jurassic_park_kitchen.jpg')
    gui = CropIDE('AirtestX IDE')
    image = Image.open('screen.png')
    gui.draw_image(image)
    gui.tag_point(100, 100)
    # gui.canvas.create_rectangle(10, 60, 30, 70, fill="red", stipple="gray12")
    gui.mainloop()


if __name__ == '__main__':
    # main()
    # atx.connect().screenshot().save('screen.png')
    atx_ide(None)
    # test()
# canvas.create_line(0, 0, 200, 100)
# canvas.create_line(0, 100, 200, 0, fill="red", dash=(4, 4))

# canvas.create_rectangle(100, 25, 150, 75)#, fill="blue")
# print (width, height)

# lastx, lasty = 0, 0
# color = 'red'
# def xy(event):
#     global lastx, lasty
#     lastx, lasty = canvas.canvasx(event.x), canvas.canvasy(event.y)


# def add_line(event):
#     global lastx, lasty
#     x, y = canvas.canvasx(event.x), canvas.canvasy(event.y)
#     canvas.create_line((lastx, lasty, x, y), fill=color, width=5, tags='boundsLine')
#     lastx, lasty = x, y

# def doneStroke(event):
#     canvas.itemconfigure('boundsLine', width=2)      
        
# canvas.bind("<Button-1>", xy)
# canvas.bind("<B1-Motion>", add_line)
# canvas.bind("<B1-ButtonRelease>", doneStroke)

# # mainframe = ttk.Frame(root, padding="3 3 12 12")
# root.mainloop()


#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import unicode_literals


import time


class CountdownTimer(object):
    def __init__(self, timeout):
        self._timeout = timeout
        self._timeend = time.time() + timeout
    
    def reset(self, timeout=None):
        if timeout:
            self._timeout = timeout
        self._timeend = time.time() + self._timeout
        
    def ticking(self):
        return time.time() < self._timeend


# coding: utf-8

import os
import sys
import logging
import webbrowser
import socket
import time
import json
import traceback
import shutil
import locale

import cv2
import tornado.ioloop
import tornado.web
import tornado.websocket
from tornado.concurrent import run_on_executor
from concurrent.futures import ThreadPoolExecutor   # `pip install futures` for python2

import atx
from atx import logutils
from atx import imutils
from atx import base


__dir__ = os.path.dirname(os.path.abspath(__file__))
log = logutils.getLogger("webide", level=logging.DEBUG)
log.setLevel(logging.DEBUG)


IMAGE_PATH = ['.', 'imgs', 'images']
device = None
atx_settings = {}
latest_screen = ''


def read_file(filename, default=''):
    if not os.path.isfile(filename):
        return default
    with open(filename, 'rb') as f:
        return f.read()

def write_file(filename, content):
    with open(filename, 'w') as f:
        f.write(content.encode('utf-8'))

def get_valid_port():
    for port in range(10010, 10100):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        result = sock.connect_ex(('127.0.0.1', port))
        sock.close()
        if result != 0:
            return port
    raise SystemError("Can not find a unused port, amazing!")


class FakeStdout(object):
    def __init__(self, fn=sys.stdout.write):
        self._fn = fn

    def write(self, s):
        self._fn(s)

    def flush(self):
        pass


class ImageHandler(tornado.web.RequestHandler):
    def get(self):
        imgs = base.list_images(path=IMAGE_PATH)
        images = []
        screenshots = []
        for name in imgs:
            realpath = name.replace('\\', '/') # fix for windows
            name = os.path.basename(name).split('@')[0]
            if realpath.startswith('screenshots/'):
                screenshots.append([name, realpath])
            else:
                images.append([name, realpath])
        self.write({
            'images': images,
            'screenshots': screenshots,
            'baseURL': self.request.protocol + '://' + self.request.host+'/static_imgs/'
        })


class MainHandler(tornado.web.RequestHandler):
    def get(self):
        imgs = base.list_images(path=IMAGE_PATH)
        imgs = [(os.path.basename(name), name) for name in imgs]
        self.render('index.html', images=imgs)

    def post(self):
        print self.get_argument('xml_text')
        self.write("Good")


class DebugWebSocket(tornado.websocket.WebSocketHandler):
    executor = ThreadPoolExecutor(max_workers=1)

    def open(self):
        log.info("WebSocket connected")
        self._run = False

    def _highlight_block(self, id):
        self.write_message({'type': 'highlight', 'id': id})
        if not self._run:
            raise RuntimeError("stopped")
        else:
            time.sleep(.1)

    def write_console(self, s):
        self.write_message({'type': 'console', 'output': s})

    def run_blockly(self, code):
        filename = '__tmp.py'
        fake_sysout = FakeStdout(self.write_console)

        __sysout = sys.stdout
        sys.stdout = fake_sysout # TODOs
        self.write_message({'type': 'console', 'output': '# '+time.strftime('%H:%M:%S') + ' start running\n'})
        try:
            # python code always UTF-8
            code = code.encode('utf-8')
            # hot patch
            code = code.replace('atx.click_image', 'd.click_image')

            exec code in {
                'highlight_block': self._highlight_block,
                '__name__': '__main__',
                '__file__': filename}
        except RuntimeError as e:
            if str(e) != 'stopped':
                raise
            print 'Program stopped'
        except Exception as e:
            self.write_message({'type': 'traceback', 'output': traceback.format_exc()})
        finally:
            self._run = False
            self.write_message({'type': 'run', 'status': 'ready'})
            sys.stdout = __sysout

    @run_on_executor
    def background_task(self, code):
        self.write_message({'type': 'run', 'status': 'running'})
        self.run_blockly(code)
        return True

    @tornado.gen.coroutine
    def on_message(self, message_text):
        message = None
        try:
            message = json.loads(message_text)
        except:
            print 'Invalid message from browser:', message_text
            return
        command = message.get('command')

        if command == 'refresh':
            imgs = base.list_images(path=IMAGE_PATH)
            images = []
            screenshots = []
            for name in imgs:
                realpath = name.replace('\\', '/') # fix for windows
                name = os.path.basename(name).split('@')[0]
                if realpath.startswith('screenshots/'):
                    screenshots.append({'name':name, 'path':realpath})
                else:
                    images.append({'name':name, 'path':realpath})
            self.write_message({'type': 'image_list', 'images': images, 'screenshots':screenshots, 'latest': latest_screen})
        elif command == 'stop':
            self._run = False
            self.write_message({'type': 'run', 'notify': '停止中'})
        elif command == 'run':
            if self._run:
                self.write_message({'type': 'run', 'notify': '运行中'})
                return
            self._run = True
            res = yield self.background_task(message.get('code'))
            self.write_message({'type': 'run', 'status': 'ready', 'notify': '运行结束', 'result': res})
        else:
            self.write_message(u"You said: " + message)

    def on_close(self):
        log.info("WebSocket closed")

    def check_origin(self, origin):
        return True


class WorkspaceHandler(tornado.web.RequestHandler):
    def get(self):
        ret = {}
        ret['xml_text'] = read_file('blockly.xml', default='<xml xmlns="http://www.w3.org/1999/xhtml"></xml>')
        ret['python_text'] = read_file('blockly.py')
        self.write(ret)

    def post(self):
        log.info("Save workspace")
        xml_text = self.get_argument('xml_text')
        python_text = self.get_argument('python_text')
        write_file('blockly.xml', xml_text)
        write_file('blockly.py', python_text)


class ScreenshotHandler(tornado.web.RequestHandler):

    def get(self):
        d = atx.connect(**atx_settings)
        v = self.get_argument('v')
        global latest_screen
        latest_screen = 'screenshots/screen_%s.png' % v
        d.screenshot(latest_screen)

        self.set_header('Content-Type', 'image/png')
        with open(latest_screen, 'rb') as f:
            while 1:
                data = f.read(16000)
                if not data:
                    break
                self.write(data)
        self.finish()

    def post(self):
        screenname = self.get_argument('screenname')
        filename = self.get_argument('filename').encode(locale.getpreferredencoding(), 'ignore')
        bound = self.get_arguments('bound[]')
        l, t, r, b = map(int, bound)
        image = imutils.open(screenname)
        image = imutils.crop(image, l, t, r, b)
        cv2.imwrite(filename, image)
        self.write({'status': 'ok'})


class StaticFileHandler(tornado.web.StaticFileHandler):
    def get(self, path=None, include_body=True):
        path = path.encode(base.SYSTEM_ENCODING) # fix for windows
        return super(StaticFileHandler, self).get(path, include_body)


def make_app(settings={}):
    static_path = os.getcwd()
    application = tornado.web.Application([
        (r"/", MainHandler),
        (r'/ws', DebugWebSocket), # code debug
        (r"/workspace", WorkspaceHandler), # save and write workspace
        (r"/images/screenshot", ScreenshotHandler),
        (r'/api/images', ImageHandler),
        (r'/static_imgs/(.*)', StaticFileHandler, {'path': static_path}),
    ], **settings)
    return application


def main(web_port=None, host=None, port=None, open_browser=True, workdir='.'):
    os.chdir(workdir)

    global IMAGE_PATH
    if not os.path.exists('screenshots'):
        os.makedirs('screenshots')
    IMAGE_PATH.append('screenshots')

    application = make_app({
        'static_path': os.path.join(__dir__, 'static'),
        'template_path': os.path.join(__dir__, 'static'),
        'debug': True,
    })
    if not web_port:
        web_port = get_valid_port()

    global atx_settings
    atx_settings['host'] = host
    atx_settings['port'] = port

    if open_browser:
        url = 'http://127.0.0.1:{}'.format(web_port)
        webbrowser.open(url, new=2) # 2: open new tab if possible

    application.listen(web_port)
    log.info("Server started.")
    log.info("Listening port on 127.0.0.1:{}".format(web_port))
    tornado.ioloop.IOLoop.instance().start()


if __name__ == "__main__":
    main()



# coding: utf-8

import subprocess32 as subprocess
import tornado
import tornado.web
import tornado.ioloop
import tornado.websocket

from tornado.concurrent import run_on_executor
from concurrent.futures import ThreadPoolExecutor   # `pip install futures` for python2


class MainHandler(tornado.web.RequestHandler):
    def get(self):
        self.render('index.html')


class TestHandler(tornado.web.RequestHandler):
    executor = ThreadPoolExecutor(max_workers=1)
    output = ''
    running = False

    @run_on_executor
    def background_test(self):
        self.running = True
        proc = subprocess.Popen('echo hello', shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        while True:
            line = proc.stdout.readline()
            if line == '':
                break
            print line
            for client in ProgressHandler.clients:
                client.write_message(line)
            self.output = self.output + line
        self.running = False

    def get(self):
        self.render('runtest.html')

    def post(self):
        if not self.running:
            self.background_test()
        self.render('runtest.html', running=self.running)


class ProgressHandler(tornado.websocket.WebSocketHandler):
    clients = set()

    def open(self):
        print 'websocket connected'
        self.write_message(TestHandler.output)
        self.clients.add(self)

    def on_close(self):
        self.clients.remove(self)
        print "WebSocket closed"
    
    def check_origin(self, origin):
        return True


def make_app(**settings):
    settings['template_path'] = 'templates'
    settings['static_path'] = 'static'
    return tornado.web.Application([
        (r"/", MainHandler),
        (r"/runtest", TestHandler),
        (r"/ws/progress", ProgressHandler),
    ], **settings)


def main():
    app = make_app(debug=True)
    app.listen(4000)
    print 'listening on port 4000'
    tornado.ioloop.IOLoop.current().start()


if __name__ == '__main__':
    main()


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
import sys
import inspect
from contextlib import contextmanager

from PIL import Image
from atx import ioskit


def inject(func, kwargs):
    """ Allow call func with more arguments """
    args = []
    for name in inspect.getargspec(func).args:
        args.append(kwargs.get(name))
    return func(*args)


def load_main(module_name):
    def _inner(parser_args):
        __import__(module_name)
        mod = sys.modules[module_name]
        pargs = vars(parser_args)
        print pargs
        return inject(mod.main, pargs)
    return _inner


def _screencap(args):
    dev = ioskit.Device(args.udid)
    image = dev.screenshot()
    if args.rotate:
        method = getattr(Image, 'ROTATE_{}'.format(args.rotate))
        image = image.transpose(method)
    image.save(args.output)
    print 'Screenshot saved to "%s"' % args.output


def main():
    ap = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    ap.add_argument("-u", "--udid", required=False, help="iPhone udid")

    subp = ap.add_subparsers()

    @contextmanager
    def add_parser(name):
        yield subp.add_parser(name, formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    
    with add_parser('developer') as p:
        p.set_defaults(func=load_main('atx.cmds.iosdeveloper'))

    with add_parser('screencap') as p:
        p.add_argument('-o', '--output', default='screenshot.png', help='take iPhone screenshot')
        p.add_argument('-r', '--rotate', type=int, choices=[0, 90, 180, 270], default=0, help='screen rotation')
        p.set_defaults(func=_screencap)

    args = ap.parse_args()
    args.func(args)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# coding: utf-8
#
# > How to get tkinter canvas to dynamically resize to window width?
# http://stackoverflow.com/questions/22835289/how-to-get-tkinter-canvas-to-dynamically-resize-to-window-width
#
# > Canvas tutoril
# http://www.tkdocs.com/tutorial/canvas.html
#
# > Canvas API reference
# http://effbot.org/tkinterbook/canvas.htm
#
# > Tutorial canvas tk
# http://www.tutorialspoint.com/python/tk_canvas.htm

import os
import time
import threading
import logging
import Tkinter as tk
import tkSimpleDialog
import tkFileDialog
from Queue import Queue

import atx
from atx import logutils, ioskit
from PIL import Image, ImageTk

log = logutils.getLogger('tkgui')
log.setLevel(logging.DEBUG)


def insert_code(filename, code, save=True, marker='# ATX CODE END'):
    """ Auto append code """
    content = ''
    found = False
    for line in open(filename, 'rb'):
        if not found and line.strip() == marker:
            found = True
            cnt = line.find(marker)
            content += line[:cnt] + code
        content += line
    if not found:
        if not content.endswith('\n'):
            content += '\n'
        content += code + marker + '\n'
    if save:
        with open(filename, 'wb') as f:
            f.write(content)
    return content

class CropIDE(object):
    def __init__(self, title='AirtestX Basic GUI', ratio=0.5, device=None):
        self._device = device
        self._root = tk.Tk()
        self._root.title(title)
        self._queue = Queue()

        self._refresh_text = tk.StringVar()
        self._refresh_text.set("Refresh")
        self._gencode_text = tk.StringVar()
        self._genfile_name = tk.StringVar()
        self._fileext_text = tk.StringVar()
        self._auto_refresh_var = tk.BooleanVar()
        self._uiauto_detect_var = tk.BooleanVar()
        self._attachfile_text = tk.StringVar()
        self._running = False # if background is running

        self._init_items()
        self._init_thread()
        self._init_refresh()

        self._lastx = 0
        self._lasty = 0
        self._bounds = None # crop area
        self._center = (0, 0) # center point
        self._offset = (0, 0) # offset to image center
        self._poffset = (0, 0)
        self._size = (90, 90)
        self._moved = False # click or click and move
        self._color = 'red' # draw color
        self._tkimage = None # keep reference
        self._image = None
        self._ratio = ratio
        self._uinodes = [] # ui dump
        self._selected_node = None
        self._hovered_node = None
        self._save_parent_dir = None

        self._init_vars()

    def _init_items(self):
        """
        .---------------.
        | Ctrl | Screen |
        |------|        |
        | Code |        |
        |      |        |
        """
        root = self._root
        root.resizable(0, 0)

        frm_control = tk.Frame(root, bg='#bbb')
        frm_control.grid(column=0, row=0, padx=5, sticky=tk.NW)
        frm_screen = tk.Frame(root, bg='#aaa')
        frm_screen.grid(column=1, row=0)

        frm_screenshot = tk.Frame(frm_control)
        frm_screenshot.grid(column=0, row=0, sticky=tk.W)
        tk.Label(frm_control, text='-'*30).grid(column=0, row=1, sticky=tk.EW)
        frm_code = tk.Frame(frm_control)
        frm_code.grid(column=0, row=2, sticky=tk.EW)

        self._btn_refresh = tk.Button(frm_screenshot, textvariable=self._refresh_text, command=self._refresh_screen)
        self._btn_refresh.grid(column=0, row=0, sticky=tk.W)
        # tk.Button(frm_screenshot, text="Wakeup", command=self._device.wakeup).grid(column=0, row=1, sticky=tk.W)
        tk.Button(frm_screenshot, text=u"保存选中区域", command=self._save_crop).grid(column=0, row=1, sticky=tk.W)
        
        # tk.Button(frm_screenshot, text="保存截屏", command=self._save_screenshot).grid(column=0, row=2, sticky=tk.W)
        frm_checkbtns = tk.Frame(frm_screenshot)
        frm_checkbtns.grid(column=0, row=3, sticky=(tk.W, tk.E))
        tk.Checkbutton(frm_checkbtns, text="Auto refresh", variable=self._auto_refresh_var, command=self._run_check_refresh).grid(column=0, row=0, sticky=tk.W)
        tk.Checkbutton(frm_checkbtns, text="UI detect", variable=self._uiauto_detect_var).grid(column=1, row=0, sticky=tk.W)

        frm_code_editor = tk.Frame(frm_code)
        frm_code_editor.grid(column=0, row=0, sticky=(tk.W, tk.E))
        tk.Label(frm_code_editor, text='Generated code').grid(column=0, row=0, sticky=tk.W)
        tk.Entry(frm_code_editor, textvariable=self._gencode_text, width=30).grid(column=0, row=1, sticky=tk.W)
        tk.Label(frm_code_editor, text='Save file name').grid(column=0, row=2, sticky=tk.W)
        tk.Entry(frm_code_editor, textvariable=self._genfile_name, width=30).grid(column=0, row=3, sticky=tk.W)
        tk.Label(frm_code_editor, text='Extention name').grid(column=0, row=4, sticky=tk.W)
        tk.Entry(frm_code_editor, textvariable=self._fileext_text, width=30).grid(column=0, row=5, sticky=tk.W)
        
        frm_code_btns = tk.Frame(frm_code)
        frm_code_btns.grid(column=0, row=2, sticky=(tk.W, tk.E))
        tk.Button(frm_code_btns, text='Run', command=self._run_code).grid(column=0, row=0, sticky=tk.W)
        self._btn_runedit = tk.Button(frm_code_btns, state=tk.DISABLED, text='Insert and Run', command=self._run_and_insert)
        self._btn_runedit.grid(column=1, row=0, sticky=tk.W)
        tk.Button(frm_code, text='Select File', command=self._run_selectfile).grid(column=0, row=4, sticky=tk.W)
        tk.Label(frm_code, textvariable=self._attachfile_text).grid(column=0, row=5, sticky=tk.W)
        tk.Button(frm_code, text='Reset', command=self._reset).grid(column=0, row=6, sticky=tk.W)

        self.canvas = tk.Canvas(frm_screen, bg="blue", bd=0, highlightthickness=0, relief='ridge')
        self.canvas.grid(column=0, row=0, padx=10, pady=10)
        self.canvas.bind("<Button-1>", self._stroke_start)
        self.canvas.bind("<B1-Motion>", self._stroke_move)
        self.canvas.bind("<B1-ButtonRelease>", self._stroke_done)
        self.canvas.bind("<Motion>", self._mouse_move)

    def _init_vars(self):
        self.draw_image(self._device.screenshot())
        if self._uiauto_detect_var.get():
            self._uinodes = self._device.dump_nodes()

    def _worker(self):
        que = self._queue
        while True:
            (func, args, kwargs) = que.get()
            try:
                func(*args, **kwargs)
            except Exception as e:
                print(e)
            finally:
                que.task_done()
    
    def _run_check_refresh(self):
        auto = self._auto_refresh_var.get()
        state = tk.DISABLED if auto else tk.NORMAL
        self._btn_refresh.config(state=state)

    def _run_async(self, func, args=(), kwargs={}):
        self._queue.put((func, args, kwargs))

    def _init_thread(self):
        th = threading.Thread(name='thread', target=self._worker)
        th.daemon = True
        th.start()

    def _init_refresh(self):
        if not self._running and self._auto_refresh_var.get():
            self._refresh_screen()
        self._root.after(200, self._init_refresh)

    def _fix_bounds(self, bounds):
        bounds = [x/self._ratio for x in bounds]
        (x0, y0, x1, y1) = bounds
        if x0 > x1:
            x0, y0, x1, y1 = x1, y1, x0, y0
        # in case of out of bounds
        w, h = self._size
        x0 = max(0, x0)
        y0 = max(0, y0)
        x1 = min(w, x1)
        y1 = min(h, y1)
        return map(int, [x0, y0, x1, y1])

    @property
    def select_bounds(self):
        if self._bounds is None:
            return None
        return self._fix_bounds(self._bounds)

    def _fix_path(self, path):
        try:
            return os.path.relpath(path, os.getcwd())
        except:
            return path

    def _save_screenshot(self):
        save_to = tkFileDialog.asksaveasfilename(**dict(
            defaultextension=".png",
            filetypes=[('PNG', '.png')],
            title='Select file'))
        if not save_to:
            return
        log.info('Save to: %s', save_to)
        self._image.save(save_to)    

    def _save_crop(self):
        log.debug('crop bounds: %s', self._bounds)
        if self._bounds is None:
            return
        bounds = self.select_bounds
        # ext = '.%dx%d.png' % tuple(self._size)
        # tkFileDialog doc: http://tkinter.unpythonic.net/wiki/tkFileDialog
        save_to = tkFileDialog.asksaveasfilename(**dict(
            initialdir=self._save_parent_dir,
            defaultextension=".png",
            filetypes=[('PNG', ".png")],
            title='Select file'))
        if not save_to:
            return
        save_to = self._fix_path(save_to)
        # force change extention with info (resolution and offset)
        save_to = os.path.splitext(save_to)[0] + self._fileext_text.get()

        self._save_parent_dir = os.path.dirname(save_to)

        log.info('Crop save to: %s', save_to)
        self._image.crop(bounds).save(save_to)
        self._genfile_name.set(os.path.basename(save_to))
        self._gencode_text.set('d.click_image(r"%s")' % save_to)

    def _run_code(self):
        d = self._device
        logging.debug("run code: %s", d)
        code = self._gencode_text.get()
        exec(code)

    def _run_and_insert(self):
        self._run_code()
        filename = self._attachfile_text.get().strip()
        code_snippet = self._gencode_text.get().strip()
        if filename and code_snippet:
            insert_code(filename, code_snippet+'\n')

    def _run_selectfile(self):
        filename = tkFileDialog.askopenfilename(**dict(
            filetypes=[('All files', '.*'), ('Python', '.py')],
            title='Select file'))
        self._attachfile_text.set(filename)
        if filename:
            self._btn_runedit.config(state=tk.NORMAL)
        print(filename)

    def _refresh_screen(self):
        def foo():
            self._running = True
            image = self._device.screenshot()
            self.draw_image(image)
            self._refresh_text.set("Refresh")

            self._draw_lines()
            self._running = False
            if self._uiauto_detect_var.get():
                self._uinodes = self._device.dump_nodes()

        self._run_async(foo)
        self._refresh_text.set("Refreshing ...")

    def _stroke_start(self, event):
        self._moved = False
        c = self.canvas
        self._lastx, self._lasty = c.canvasx(event.x), c.canvasy(event.y)
        log.debug('mouse position: %s', (self._lastx, self._lasty))

    def _stroke_move(self, event):
        self._moved = True
        self._reset()
        c = self.canvas
        x, y = c.canvasx(event.x), c.canvasy(event.y)
        self._bounds = (self._lastx, self._lasty, x, y)
        self._center = (self._lastx+x)/2, (self._lasty+y)/2
        self._draw_lines()

    def _stroke_done(self, event):
        c = self.canvas
        x, y = c.canvasx(event.x), c.canvasy(event.y)
        if self._moved: # drag action
            x, y = (self._lastx+x)/2, (self._lasty+y)/2
            self._offset = (0, 0)
        else:
            # click action
            if self._bounds is None:
                cx, cy = (x/self._ratio, y/self._ratio)
                if self._uiauto_detect_var.get() and self._hovered_node:
                    self._selected_node = self._hovered_node
                    log.debug("select node: %s", repr(self._selected_node))
                    log.debug("center: %s", self._selected_node.bounds.center)
                    # self._device.click(cx, cy)

                self._gencode_text.set('d.click(%d, %d)' % (cx, cy))
            else:
                (x0, y0, x1, y1) = self.select_bounds
                ww, hh = x1-x0, y1-y0
                cx, cy = (x/self._ratio, y/self._ratio)
                mx, my = (x0+x1)/2, (y0+y1)/2 # middle
                self._offset = (offx, offy) = map(int, (cx-mx, cy-my))
                poffx = ww and round(offx*100.0/ww) # in case of ww == 0
                poffy = hh and round(offy*100.0/hh)
                self._poffset = (poffx, poffy)
                self._gencode_text.set('(%d, %d)' % (cx, cy)) #offset=(%.2f, %.2f)' % (poffx/100, poffy/100))
                # self._gencode_text.set('offset=(%.2f, %.2f)' % (poffx/100, poffy/100))

        ext = ".%dx%d" % tuple(self._size)
        if self._poffset != (0, 0):
            px, py = self._poffset
            ext += '.%s%d%s%d' % (
                'R' if px > 0 else 'L', abs(px), 'B' if py > 0 else 'T', abs(py))
        ext += '.png'
        self._fileext_text.set(ext)
        self._center = (x, y) # rember position
        self._draw_lines()
        self.canvas.itemconfigure('select-bounds', width=2)

    def draw_image(self, image):
        self._image = image
        self._size = (width, height) = image.size
        w, h = int(width*self._ratio), int(height*self._ratio)
        image = image.copy()
        image.thumbnail((w, h), Image.ANTIALIAS)
        tkimage = ImageTk.PhotoImage(image)
        self._tkimage = tkimage # keep a reference
        self.canvas.config(width=w, height=h)
        self.canvas.create_image(0, 0, anchor=tk.NW, image=tkimage)

    def _draw_bounds(self, bounds, color=None, tags='select-bounds'):
        if not color:
            color=self._color
        c = self.canvas
        (x0, y0, x1, y1) = bounds
        c.create_rectangle(x0, y0, x1, y1, outline=color, tags='select-bounds', width=2)

    def _draw_lines(self):
        if self._center and self._center != (0, 0):
            x, y = self._center
            self.draw_point(x, y)
        if self._bounds:
            self._draw_bounds(self._bounds)
        if self._hovered_node:
            # print self._hovered_node.bounds
            bounds = [v*self._ratio for v in self._hovered_node.bounds]
            self._draw_bounds(bounds, color='blue', tags='ui-bounds')

    def _reset(self):
        self._bounds = None
        self._offset = (0, 0)
        self._poffset = (0, 0)
        self._center = (0, 0)
        self.canvas.delete('select-bounds')
        self.canvas.delete('select-point')
        self.canvas.delete('ui-bounds')

    def _mouse_move(self, event):
        if not self._uiauto_detect_var.get():
            return
        c = self.canvas
        x, y = c.canvasx(event.x), c.canvasy(event.y)
        x, y = x/self._ratio, y/self._ratio
        # print x, y
        hovered_node = None
        min_area = None
        for node in self._uinodes:
            if node.bounds.is_inside(x, y):
                if min_area is None or node.bounds.area < min_area:
                    hovered_node = node
                    min_area = node.bounds.area
        if hovered_node:
            self._hovered_node = hovered_node
            self._reset()
            self._draw_lines()

    def draw_point(self, x, y):
        self.canvas.delete('select-point')
        r = max(min(self._size)/30*self._ratio, 5)
        self.canvas.create_line(x-r, y, x+r, y, width=2, fill=self._color, tags='select-point')
        self.canvas.create_line(x, y-r, x, y+r, width=2, fill=self._color, tags='select-point')

    def mainloop(self):
        self._root.mainloop()
        

def main(serial, host='127.0.0.1', scale=0.5, platform=None):
    log.debug("gui starting(scale: {}) ...".format(scale))
    if platform == 'android':
        d = atx.connect(serial, host=host, platform='android')
        serial = d.serial
    elif platform == 'ios':
        d = atx.connect(serial, platform='ios')
        serial = serial
    else:
        d = atx.connect(serial)
        serial = serial
        # d = ioskit.Device(udid=serial)
        # d = atx.connect(udid=serial, platform='ios')
        # serial = d.udid
    gui = CropIDE('ATX GUI SN: %s' % serial, ratio=scale, device=d)
    gui.mainloop()

def test():
    # image = Image.open('jurassic_park_kitchen.jpg')
    gui = CropIDE('AirtestX IDE')
    image = Image.open('screen.png')
    gui.draw_image(image)
    gui.draw_point(100, 100)
    gui.mainloop()


if __name__ == '__main__':
    main(None, platform='ios')
    # main()
    # atx.connect().screenshot().save('screen.png')
    # test()



#!/usr/bin/python
# This is a simple port-forward / proxy, written using only the default python
# library. If you want to make a suggestion or fix something you can contact-me
# at voorloop_at_gmail.com
# Distributed over IDC(I Don't Care) license
import socket
import select
import time
import sys

# Changing the buffer_size and delay, you can improve the speed and bandwidth.
# But when buffer get to high or delay go too down, you can broke things
buffer_size = 4096
delay = 0.0001
forward_to = ('127.0.0.1', 5037)
#forward_to = ('10.249.80.122', 57149)

class Forward:
    def __init__(self):
        self.forward = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    def start(self, host, port):
        try:
            self.forward.connect((host, port))
            return self.forward
        except Exception, e:
            print e
            return False

class TheServer:
    input_list = []
    channel = {}

    def __init__(self, host, port):
        self.server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.server.bind((host, port))
        self.server.listen(200)

    def main_loop(self):
        self.input_list.append(self.server)
        while 1:
            time.sleep(delay)
            ss = select.select
            inputready, outputready, exceptready = ss(self.input_list, [], [])
            for self.s in inputready:
                if self.s == self.server:
                    self.on_accept()
                    break

                self.data = self.s.recv(buffer_size)
                if len(self.data) == 0:
                    self.on_close()
                    break
                else:
                    self.on_recv()

    def on_accept(self):
        forward = Forward().start(forward_to[0], forward_to[1])
        clientsock, clientaddr = self.server.accept()
        if forward:
            print '[proxy]', clientaddr, "has connected"
            self.input_list.append(clientsock)
            self.input_list.append(forward)
            self.channel[clientsock] = forward
            self.channel[forward] = clientsock
        else:
            print "[proxy] Can't establish connection with remote server.",
            print "[proxy] Closing connection with client side", clientaddr
            clientsock.close()

    def on_close(self):
        print '[proxy]', self.s.getpeername(), "has disconnected"
        print '[proxy]', self.channel[self.s].getpeername(), "has disconnected, too"
        #remove objects from input_list
        self.input_list.remove(self.s)
        self.input_list.remove(self.channel[self.s])
        out = self.channel[self.s]
        # close the connection with client
        self.channel[out].close()  # equivalent to do self.s.close()
        # close the connection with remote server
        self.channel[self.s].close()
        # delete both objects from channel dict
        del self.channel[out]
        del self.channel[self.s]

    def on_recv(self):
        data = self.data
        # here we can parse and/or modify the data before send forward
        #print isinstance(self.s, Forward), self.server
        print '[%5s]: %s' % (self.s.getpeername()[1], data)
        self.channel[self.s].send(data)

if __name__ == '__main__':
        server = TheServer('', 5555)
        try:
            server.main_loop()
        except KeyboardInterrupt:
            print "Ctrl C - Stopping server"
            sys.exit(1)



#-*- encoding: utf-8 -*-

import os
import cv2
import yaml
import numpy as np
from collections import defaultdict

def find_match(img, tmpl, rect=None, mask=None):
    if rect is not None:
        h, w = img.shape[:2]
        x, y, x1, y1 = rect
        if x1 > w or y1 > h:
            return 0, None
        img = img[y:y1, x:x1, :]

        if mask is not None:
            img = img.copy()
            img[mask!=0] = 0
            tmpl = tmpl.copy()
            tmpl[mask!=0] = 0

    s_bgr = cv2.split(tmpl) # Blue Green Red
    i_bgr = cv2.split(img)

    weight = (0.3, 0.3, 0.4)
    resbgr = [0, 0, 0]
    for i in range(3): # bgr
        resbgr[i] = cv2.matchTemplate(i_bgr[i], s_bgr[i], cv2.TM_CCOEFF_NORMED)
    match = resbgr[0]*weight[0] + resbgr[1]*weight[1] + resbgr[2]*weight[2]
    min_val, max_val, min_loc, max_loc = cv2.minMaxLoc(match)
    confidence = max_val
    x, y = max_loc
    h, w = tmpl.shape[:2]
    if rect is None:
        rect = (x, y, x+w, y+h)
    # cv2.rectangle(img, (x,y), (x+w,y+h), (0,255,0) ,2)
    # cv2.imshow('test', img)
    # cv2.waitKey(20)
    return confidence, rect

def get_mask(img1, img2, thresh=20):
    if img1.shape != img2.shape:
        return
    diff = cv2.absdiff(img1, img2)
    diff = np.mean(diff, axis=2)
    diff[diff<=thresh] = 0
    diff[diff>thresh] = 255
    mask = np.dstack([diff]*3)
    return mask

def get_match_confidence(img1, img2, mask=None):
    if img1.shape != img2.shape:
        return False
    ## first try, using absdiff
    # diff = cv2.absdiff(img1, img2)
    # h, w, d = diff.shape
    # total = h*w*d
    # num = (diff<20).sum()
    # print 'is_match', total, num
    # return num > total*0.90
    if mask is not None:
        img1 = img1.copy()
        img1[mask!=0] = 0
        img2 = img2.copy()
        img2[mask!=0] = 0
    ## using match
    match = cv2.matchTemplate(img1, img2, cv2.TM_CCOEFF_NORMED)
    _, confidence, _, _ = cv2.minMaxLoc(match)
    # print confidence
    return confidence

class SceneDetector(object):
    '''detect game scene from screen image'''

    def __init__(self, scene_directory):
        self.scene_touches = {}
        self.scene_directory = scene_directory
        self.build_tree(scene_directory)

    def build_tree(self, directory):
        '''build scene tree from images'''
        confile = os.path.join(directory, 'config.yml')
        conf = {}
        if os.path.exists(confile):
            conf = yaml.load(open(confile).read())

        class node(defaultdict):
            name = ''
            parent = None
            tmpl = None
            rect = None
            mask = None

            def __str__(self):
                obj = self
                names = []
                while obj.parent is not None:
                    names.append(obj.name)
                    obj = obj.parent
                return '-'.join(names[::-1])

        def tree():
            return node(tree)

        root = tree()
        for s in os.listdir(directory):
            if not s.endswith('.png') or s.endswith('_mask.png'): 
                continue
            obj = root
            for i in s[:-4].split('-'):
                obj[i].name = i
                obj[i].parent = obj
                obj = obj[i]
            obj.tmpl = cv2.imread(os.path.join(directory, s))
            obj.rect = conf.get(s[:-4], {}).get('rect')
            maskimg = conf.get(s[:-4], {}).get('mask')
            if maskimg is not None:
                maskimg = os.path.join(directory, maskimg)
                if os.path.exists(maskimg):
                    obj.mask = cv2.imread(maskimg)

        self.tree = root
        self.current_scene = []
        self.confile = confile
        self.conf = conf

    def match_child(self, img, node):
        c, s, r = (0, None, None)
        for scene in node.itervalues():
            if scene.tmpl is None:
                continue
            print str(scene), scene.rect, img.shape
            confidence, rect = find_match(img, scene.tmpl, scene.rect, scene.mask)
            # print scene.name, confidence, rect
            if confidence > c:
                c, s, r = (confidence, scene, rect)

        if c > 0.95:
            key = str(s)
            if key not in self.conf:
                self.conf[key] = {}

            changed = False
            if c > self.conf[key].get('confidence', 0):
                s.rect = r
                self.conf[key]['confidence'] = c
                self.conf[key]['rect'] = list(r)
                changed = True

            if changed or s.mask is None:
                x, y, x1, y1 = r
                s.mask = get_mask(img[y:y1, x:x1, :], s.tmpl, 20)
                maskimg = os.path.join(self.scene_directory, '%s_mask.png' % key)
                cv2.imwrite(maskimg, s.mask)
                self.conf[key]['mask'] = maskimg
                changed = True

            if changed:
                self.save_config()

        return c, s, r

    def save_config(self):
        print 'save config', self.conf
        with open(self.confile, 'w') as f:
            yaml.dump(self.conf, f)

    def detect(self, img):
        # check current scene path        
        # print 'checking current scene'
        if self.current_scene:
            for i in range(len(self.current_scene)):
                s, r = self.current_scene[i]
                x, y, x1, y1 = r
                c = get_match_confidence(img[y:y1, x:x1, :], s.tmpl, s.mask)
                if c < 0.75:
                    break
            else:
                # print 'current scene ok'
                s = self.current_scene[-1][0]
                if len(s.values()) == 0:
                    return s
            self.current_scene = self.current_scene[:i]

        # top scene has changed
        if not self.current_scene:
            c, s, r = self.match_child(img, self.tree)
            if c < 0.75:
                return
            self.current_scene = [(s, r)]
        
        s = self.current_scene[-1][0]
        while True:
            c, s, r = self.match_child(img, s)
            if c < 0.75:
                break
            self.current_scene.append((s, r))

        return s


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# License under MIT


from __future__ import absolute_import

import os
import json
import time

import yaml
import subprocess32 as subprocess
from PIL import Image

from atx import consts
from atx import errors
from atx import patch
from atx import base
from atx import imutils
from atx import strutils
from atx import logutils
from atx import ioskit
from atx.drivers import Bounds, Display
from atx.drivers.mixin import DeviceMixin, hook_wrap


__dir__ = os.path.dirname(os.path.abspath(__file__))
log = logutils.getLogger(__name__)

class IOSDevice(DeviceMixin):
    def __init__(self, bundle_id=None, udid=None):
        DeviceMixin.__init__(self)

        self.d = ioskit.Device(udid)
        self.udid = self.d.udid

        self._proc = None
        self._display = None #Display(2208, 1242)
        self._scale = 1
        self._env = os.environ.copy()
        self._init_display()

        self.screen_rotation = 1 # TODO: auto judge

        if not bundle_id:
            print 'WARNING [ios.py]: bundle_id is not set' #, use "com.netease.atx.apple" instead.'
            # self._init_instruments('com.netease.atx.apple')
        else:
            self._init_instruments(bundle_id)

    def _init_display(self):
        model = self.d.info['HardwareModel']
        with open(os.path.join(__dir__, 'ios-models.yml'), 'rb') as f:
            items = yaml.load(f.read())
        for item in items:
            if model == item.get('model'):
                (width, height) = map(int, item.get('pixel').split('x'))
                self._scale = item.get('scale')
                self._display = Display(width*self._scale, height*self._scale)
                break
        if self._display is None:
            raise RuntimeError("TODO: not support your phone for now, You need contact the author.")
    
    def _init_instruments(self, bundle_id):
        self._bootstrap = os.path.join(__dir__, 'bootstrap.sh')
        self._bundle_id = bundle_id
        self._env.update({'UDID': self.udid, 'BUNDLE_ID': self._bundle_id})
        # 1. remove pipe
        # subprocess.check_output([self._bootstrap, 'reset'], env=self._env)
        # 2. start instruments
        self._proc = subprocess.Popen([self._bootstrap, 'instruments'], env=self._env, stdout=subprocess.PIPE)
        self.sleep(5.0)
        self._wait_instruments()

    def _wait_instruments(self):
        ret = self._run('1')
        if ret != 1:
            log.error('Instruments stdout:\n' + self._proc.stdout.read())
            raise RuntimeError('Instruments start failed, expect 1 but got %s' % (ret,))

    def _run(self, code):
        # print self._proc.poll()
        # print code
        encoded_code = json.dumps({'command': code})
        output = subprocess.check_output([self._bootstrap, 'run', encoded_code], env=self._env)
        # print output
        try:
            return json.loads(output)
        except:
            print 'unknown json output:', output
            return output

    def _run_nowait(self, code):
        ''' TODO: change to no wait '''
        print self._proc.poll()
        encoded_code = json.dumps({'command': code, 'nowait': True})
        output = subprocess.check_output([self._bootstrap, 'run', '--nowait', encoded_code], env=self._env)
        return output

    def _close(self):
        print 'Terminate instruments'
        if self._proc:
            self._proc.terminate()
        # 1. remove pipe
        subprocess.check_output([self._bootstrap, 'reset'], env=self._env)

    def __del__(self):
        if hasattr(self, '_bootstrap'):
            self._close()

    @property
    def rotation(self):
        return self.screen_rotation
    
    @property
    def display(self):
        return self._display

    @property
    def info(self):
        return self.d.info

    def screenshot(self, filename=None):
        '''
        Take ios screenshot
        Args:
            - filename(string): optional
        Returns:
            PIL.Image object
        '''
        image = self.d.screenshot()
        if self.rotation:
            method = getattr(Image, 'ROTATE_{}'.format(self.rotation*90))
            image = image.transpose(method)
        if filename:
            image.save(filename)
        return image

    def click(self, x, y):
        '''
        Simulate click operation
        Args:
            - x (int): position of x
            - y (int): position of y
        Returns:
            self
        '''
        self._run_nowait('target.tap({x: %d, y: %d})' % (x/self._scale, y/self._scale))
        return self

    def install(self, filepath):
        self.d.install(filepath)

    def sleep(self, sec):
        self.delay(sec)

    def type(self, text):
        self._run_nowait('$.typeString(%s)' % json.dumps(text))

    def start_app(self, bundle_id):
        self.d.start_app(bundle_id)

    def current_app(self):
        ''' todo, maybe return dict is a better way '''
        return self._run('target.frontMostApp().bundleID()').strip().strip('"')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# hzsunshx <2016-06-12>
# make sure only one phone connected to pc/

import time

import atx
from atx.ext.gt import GT


d = atx.connect()

def test_gt():
    gt = GT(d)
    gt.start_test('com.netease.my')
    print 'test started. wait 5s'
    time.sleep(10.0)
    gt.stop_and_save()
    print 'save perf data'
    # time.sleep(3.0)
    # gt.quit()


if __name__ == '__main__':
    test_gt()


# coding: utf-8
#
# author: codeskyblue(hzsunshx)
# created: 2016-05-16
# modified: 2016-08-19
#
# Experiment of caculate FPS of android without root
# code reference from chromiumwebapps/chromium
#
# It is insteresting to know this way which can calculate not only fps, but also every frame use time.
# Is it really useful, I don't know, really don't know.
# so code just stop here.

import time
import subprocess
import re

nanoseconds_per_second = 1e9

# view = 'SurfaceView'

def get_top_view():
    out = subprocess.check_output(['adb', 'shell', 'dumpsys', 'SurfaceFlinger'])
    lines = out.replace('\r\n', '\n').splitlines()
    max_area = 0
    top_view = None
    for index, line in enumerate(lines):
        line = line.strip()
        if not line.startswith('+ Layer '):
            continue
        m = re.search(r'\((.+)\)', line)
        if not m:
            continue
        view_name = m.group(1)
        (x0, y0, x1, y1) = map(int, re.findall(r'(\d+)', lines[index+4]))
        cur_area = (x1-x0) * (y1-y0)
        if cur_area > max_area:
            max_area = cur_area
            top_view = view_name
    return top_view


def init_frame_data(view):
    out = subprocess.check_output(['adb', 'shell', 'dumpsys', 'SurfaceFlinger', '--latency-clear', view])
    if out.strip() != '':
        raise RuntimeError("Not supported.")
    time.sleep(0.1)
    (refresh_period, timestamps) = frame_data(view)
    base_timestamp = 0
    base_index = 0
    for timestamp in timestamps:
        if timestamp != 0:
            base_timestamp = timestamp
            break
        base_index += 1

    if base_timestamp == 0:
        raise RuntimeError("Initial frame collect failed")
    return (refresh_period, base_timestamp, timestamps[base_index:])


def frame_data(view):
    out = subprocess.check_output(['adb', 'shell', 'dumpsys', 'SurfaceFlinger', '--latency', view])
    results = out.splitlines()
    refresh_period = long(results[0]) / nanoseconds_per_second
    timestamps = []
    for line in results[1:]:
        fields = line.split()
        if len(fields) != 3:
            continue
        (start, submitting, submitted) = map(int, fields)
        if submitting == 0:
            continue

        timestamp = submitting/nanoseconds_per_second
        timestamps.append(timestamp)
    return (refresh_period, timestamps)


def continue_collect_frame_data():
    view = get_top_view()
    if view is None:
        raise RuntimeError("Fail to get current SurfaceFlinger view")
    print 'Current view:', view

    refresh_period, base_timestamp, timestamps = init_frame_data(view)
    while True:
        refresh_period, tss = frame_data(view)
        last_index = 0
        if timestamps:
            recent_timestamp = timestamps[-2]
            last_index = tss.index(recent_timestamp)
        timestamps = timestamps[:-2] + tss[last_index:]
        
        time.sleep(1.5)
        
        ajusted_timestamps = []
        for seconds in timestamps[:]:
            seconds -= base_timestamp
            if seconds > 1e6: # too large, just ignore
                continue
            ajusted_timestamps.append(seconds)

        from_time = ajusted_timestamps[-1] - 1.0
        fps_count = 0
        for seconds in ajusted_timestamps:
            if seconds > from_time:
                fps_count += 1
        print fps_count, len(ajusted_timestamps)

        # print ajusted_timestamps


if __name__ == '__main__':
    continue_collect_frame_data()



# coding: utf-8

from __future__ import absolute_import

import re
import collections

import six
from atx import imutils
from atx import strutils


FindPoint = collections.namedtuple('FindPoint', ['pos', 'confidence', 'method', 'matched'])
Display = collections.namedtuple('Display', ['width', 'height'])


__boundstuple = collections.namedtuple('Bounds', ['left', 'top', 'right', 'bottom'])
class Bounds(__boundstuple):
    def __init__(self, *args, **kwargs):
        self._area = None

    def is_inside(self, x, y):
        v = self
        return x > v.left and x < v.right and y > v.top and y < v.bottom

    @property
    def area(self):
        if not self._area:
            v = self
            self._area = (v.right-v.left) * (v.bottom-v.top)
        return self._area

    @property
    def center(self):
        v = self
        return (v.left+v.right)/2, (v.top+v.bottom)/2

    def __mul__(self, mul):
        return Bounds(*(int(v*mul) for v in self))

class ImageCrop(object):
    def __init__(self, src, bound):
        self.src = src
        l, t, w, h = bound
        self.bound = Bounds(l, t, l+w, t+h)

class Pattern(object):
    def __init__(self, name, image=None, offset=None, anchor=0, rsl=None, resolution=None, th=None, threshold=None):
        """
        Args:
            name: image filename
            image: opencv image object
            offset: offset of image center
            anchor: not supported
            resolution: image origin screen resolution
            rsl: alias of resolution
            threshold: image match threshold, usally (0, 1]
            th: alias of threshold
        """
        if isinstance(name, ImageCrop):
            self._name = name.src
            self._bound = name.bound
        else:
            self._name = name
            self._bound = None

        self._image = image # if image is None, it will delay to pattern_open function
        self._offset = offset
        self._resolution = rsl or resolution
        self._threshold = th or threshold
        if isinstance(image, six.string_types):
            self._name = image

        # search format name.1080x1920.png
        if self._resolution is None:
            m = re.search(r'\.(\d+)x(\d+)\.', self._name)
            if m:
                (w, h) = sorted(map(int, (m.group(1), m.group(2))))
                # TODO(ssx): gcd(w, h), make sure the biggest < 20
                self._resolution = (w, h)

        if self._offset is None:
            m = re.search(r'\.([LRTB])(\d+)([LRTB])(\d+)\.', self._name)
            if m:
                offx, offy = 0, 0
                for i in (1, 3):
                    flag, number = m.group(i), int(m.group(i+1))
                    if flag in ('L', 'R'):
                        offx = number/100.0 * (1 if flag == 'R' else -1)
                    if flag in ('T', 'B'):
                        offy = number/100.0 * (1 if flag == 'B' else -1)
                self._offset = (offx, offy)

    def __str__(self):
        return 'Pattern(name: {}, offset: {})'.format(strutils.encode(self._name), self.offset)

    def save(self, path):
        """ save image to path """
        import cv2
        cv2.imwrite(path, self._image)
    
    @property
    def image(self):
        if self._bound is None:
            return self._image
        else:
            return imutils.crop(self._image, *self._bound)

    @property
    def offset(self):
        return self._offset

    @property
    def resolution(self):
        return self._resolution

    @property
    def threshold(self):
        return self._threshold



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys
import subprocess
import random
import string
import time
import logging
import threading

from atx import strutils


random.seed(time.time())

def id_generator(n=5):
    return ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(n))


def dirname(name):
    if os.path.isabs(name):
        return os.path.dirname(name)
    return os.path.dirname(os.path.abspath(name))


def exec_cmd(*cmds, **kwargs):
    '''
    @arguments env=None, timeout=3
    may raise Error
    '''
    env = os.environ.copy()
    env.update(kwargs.get('env', {}))
    envcopy = {}
    for key in env: 
        try:
            envcopy[key] = str(env[key]).encode('utf-8') # fix encoding
        except:
            print('IGNORE BAD ENV KEY: ' + repr(key))
    env = envcopy

    timeout = kwargs.get('timeout', 120)
    shell = kwargs.get('shell', False)
    try:
        import sh
        # log.debug('RUN(timeout=%ds): %s'%(timeout, ' '.join(cmds)))
        if shell:
            cmds = list(cmds)
            cmds[:0] = ['bash', '-c']
        c = sh.Command(cmds[0])
        try:
            r = c(*cmds[1:], _err_to_out=True, _out=sys.stdout, _env=env, _timeout=timeout)
        except:
            # log.error('EXEC_CMD error, cmd: %s'%(' '.join(cmds)))
            raise
    except ImportError:
        # log.debug('RUN(timeout=XX): %s'%(' '.join(cmds)))
        if shell:
            cmds = ' '.join(cmds)
        r = subprocess.Popen(cmds, env=env, stdout=sys.stdout, stderr=sys.stderr, shell=shell)
        return r.wait()
    return 0


def random_name(name):
    out = []
    for c in name:
        if c == 'X':
            c = random.choice(string.ascii_lowercase)
        out.append(c)
    return ''.join(out)


def remove_force(name):
    if not os.path.isfile(name):
        return
    try:
        os.unlink(name)
    except Exception as e:
        print("Warning: tempfile {} not deleted, Error {}".format(name, e))


SYSTEM_ENCODING = 'gbk' if os.name == 'nt' else 'utf-8'
VALID_IMAGE_EXTS = ['.jpg', '.jpeg', '.png', '.bmp']


# def auto_decode(s, encoding='utf-8'):
#     return s if isinstance(s, unicode) else unicode(s, encoding)


def list_images(path=['.']):
    """ Return list of image files """
    for image_dir in set(path):
        if not os.path.isdir(image_dir):
            continue
        for filename in os.listdir(image_dir):
            bname, ext = os.path.splitext(filename)
            if ext.lower() not in VALID_IMAGE_EXTS:
                continue

            filepath = os.path.join(image_dir, filename)
            yield strutils.decode(filepath)


def list_all_image(path, valid_exts=VALID_IMAGE_EXTS):
    """List all images under path

    @return unicode list
    """
    for filename in os.listdir(path):
        bname, ext = os.path.splitext(filename)
        if ext.lower() not in VALID_IMAGE_EXTS:
            continue
        filepath = os.path.join(path, filename)
        yield strutils.decode(filepath)


def image_name_match(name, target):
    if name == target:
        return True

    bn = os.path.normpath(name)
    bt = os.path.basename(target)

    if bn == bt:
        return True

    bn, ext = os.path.splitext(bn)
    if ext != '':
        return False

    for ext in VALID_IMAGE_EXTS:
        if bn+ext == bt or bn+ext.upper() == bt:
            return True

    if bt.find('@') != -1:
        if bn == bt[:bt.find('@')]:
            return True
    return False


def search_image(name=None, path=['.']):
    """
    look for the image real path, if name is None, then return all images under path.

    @return system encoded path string
    FIXME(ssx): this code is just looking wired.
    """
    name = strutils.decode(name)
    for image_dir in path:
        if not os.path.isdir(image_dir):
            continue
        image_dir = strutils.decode(image_dir)
        image_path = os.path.join(image_dir, name)
        if os.path.isfile(image_path):
            return strutils.encode(image_path)

        for image_path in list_all_image(image_dir):
            if not image_name_match(name, image_path):
                continue
            return strutils.encode(image_path)
    return None

def clean_path(filepath):
    return os.path.normpath(os.path.relpath(strutils.decode(filepath)))


def filename_match(fsearch, filename, width, height):
    '''
    <nickname>@({width}x{height}|auto).(png|jpg|jpeg)
    '''
    fsearch = clean_path(fsearch)
    filename = clean_path(filename)
    if fsearch == filename:
        return True

    if fsearch.find('@') == -1:
        return False
    basename, fileext = os.path.splitext(fsearch)
    nickname, extinfo = basename.split('@', 1)
    if extinfo == 'auto':
        valid_names = {}.fromkeys([
            nickname+'@{}x{}'.format(width, height)+fileext,
            nickname+'@{}x{}'.format(height, width)+fileext,
            nickname+'.{}x{}'.format(width, height)+fileext,
            nickname+'.{}x{}'.format(height, width)+fileext,
        ])
        if filename in valid_names:
            return True
    # if extinfo.find('x') != -1:
    #     cw, ch = extinfo.split('x', 1)
    #     if cw*width == ch*height or cw*height == ch*width:
    #         return True
    return False


def lookup_image(fsearch, width=0, height=0):
    dirname = os.path.dirname(fsearch) or "."
    for file in os.listdir(dirname):
        filepath = os.path.join(dirname, file)
        if filename_match(fsearch, filepath, width, height):
            return filepath


def nameddict(name, props):
    """
    Point = nameddict('Point', ['x', 'y'])
    pt = Point(x=1, y=2)
    pt.y = 3
    print pt
    """
    class NamedDict(object):
        def __init__(self, *args, **kwargs):
            self.__store = {}.fromkeys(props)
            if args:
                for i, k in enumerate(props[:len(args)]):
                    self[k] = args[i]
            for k, v in kwargs.items():
                self[k] = v

        def __getattr__(self, key):
            # print '+', key
            if key.startswith('_NamedDict__'):
                return self.__dict__[key]
            else:
                return self.__store[key]

        def __setattr__(self, key, value):
            # print '-', key
            if key.startswith('_NamedDict__'):
                object.__setattr__(self, key, value)
            else:
                self.__setitem__(key, value)
                # self.__store[key] = value

        def __getitem__(self, key):
            return self.__store[key]

        def __setitem__(self, key, value):
            if key not in props:
                raise AttributeError("NamedDict(%s) has no attribute %s, avaliables are %s" % (
                    name, key, props))
            self.__store[key] = value

        def __dict__(self):
            return self.__store

        def __str__(self):
            return 'NamedDict(%s: %s)' % (name, str(self.__store))

    return NamedDict


if __name__ == '__main__':
    # print search_image('你好.png')
    # print search_image('oo')
    # print search_image()

    Point = nameddict('Point', ['x', 'y'])
    print(Point(2, 3, 4))



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import atexit
import os
import subprocess
import tempfile
import logging
import shutil

from atx.cmds.utils import http_download


__alias = {
    '9.3': '9.3 (13E230)', # 2016-05-04
}

# Can also be found in directory
IMAGE_BASE_DIR = '/Applications/Xcode.app/Contents/Developer/Platforms/iPhoneOS.platform/DeviceSupport/'
IMAGE_BASE_URL = 'http://gohttp.nie.netease.com/tools/tools-ios/DeveloperImages/'
logger = logging.getLogger('ios')


def init():
    ch = logging.StreamHandler()
    fmt = "%(asctime)s [%(name)s:%(lineno)4s] %(message)s"
    formatter = logging.Formatter(fmt)
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    logger.setLevel(logging.DEBUG)


def check_output(cmds, shell=False):
    try:
        output = subprocess.check_output(cmds, stderr=subprocess.STDOUT, shell=shell)
        return output
    except subprocess.CalledProcessError as e:
        logger.warn('Failed to run command: %s', ' '.join(cmds))
        logger.warn('Error output:\n%s', e.output)
        raise


def look_path(name, search_paths=[], env_path=True):
    os.pathsep
    if env_path:
        search_paths += os.getenv('PATH').split(os.pathsep)
    for directory in search_paths:
        if not os.path.isdir(directory):
            continue
        filepath = os.path.join(directory, name)
        if os.path.isfile(filepath):
            return filepath
    return None


__execpath = {}

def look_exec(name):
    if __execpath.get(name):
        return __execpath[name]

    ext = '.exe' if os.name == 'nt' else ''
    search_paths = [
        r'C:\Program Files (x86)\Quamotion\iMobileDevice',
        r'D:\Program Files (x86)\Quamotion\iMobileDevice',
        r'E:\Program Files (x86)\Quamotion\iMobileDevice',
        r'C:\Program Files\Quamotion\iMobileDevice',
        r'D:\Program Files\Quamotion\iMobileDevice',
        r'E:\Program Files\Quamotion\iMobileDevice',
    ]
    filepath = look_path(name+ext, search_paths)
    __execpath[name] = filepath
    return filepath


def idevice(name, *args):
    exec_name = 'idevice' + name
    exec_path = look_exec(exec_name)
    if not exec_path:
        raise EnvironmentError('Necessary binary ("%s") not found.' % exec_name)
    return check_output([exec_path] + list(args))


def devices():
    udids = [udid.strip() for udid in idevice('_id', '-l').splitlines() if udid.strip()]
    return {udid: idevice('name', '-u', udid).decode('utf-8').strip() for udid in udids}


def device_product_version(udid):
    return idevice('info', '-u', udid, '-k', 'ProductVersion').strip()


def download(filename, tmpdir, version, base_url=IMAGE_BASE_URL):
    if sys.platform == 'darwin':
        version = __alias.get(version, version)
        abs_path = os.path.join(IMAGE_BASE_DIR, version, filename)
        if os.path.exists(abs_path):
            return abs_path
    target_path = os.path.join(tmpdir, filename)
    source_url = base_url + '/'.join([version, filename])
    logger.info("Download %s/%s", version, filename)
    return http_download(source_url, target_path)


def select_device():
    devs = devices()
    if len(devs) == 0:
        raise EnvironmentError('iPhone device is not attached.')
    if len(devs) > 1:
        raise EnvironmentError('More than one iPhone device detected.')

    udid = devs.keys()[0]
    devname = devs[udid]
    return udid, devname


def is_mounted(udid):
    return 'ImagePresent: true' in idevice('imagemounter', '-l', '-u', udid)


def mount_image(udid, image_file, image_signature_file):
    if is_mounted(udid):
        logger.info("Developer image has already mounted.")
        return
    idevice('imagemounter', '-u', udid, image_file, image_signature_file)    
    logger.info("^_^ Developer image has been mounted.")


def check_enviroment():
    return look_exec('idevice_id') is not None
        

def main(udid=None):
    init()

    if not check_enviroment():
        sys.exit("No imobiledevice found in $PATH, but you can download from here\n\n    %s" %(
            "http://quamotion.mobi/iMobileDevice/Download",))

    logger.info("Make tmp dir ...")
    tmpdir = tempfile.mkdtemp(prefix='atx-ios-developer-')
    if not tmpdir:
        logger.warn("tmpdir create failed.")
        sys.exit(1)

    @atexit.register
    def _clean():
        logger.info("Cleaning tmp dir: %s", tmpdir)
        shutil.rmtree(tmpdir)

    udid, devname = select_device()
    long_version = device_product_version(udid)
    version = '.'.join(long_version.split('.')[:2])
    logger.info("Device udid is %s", udid)
    logger.info("Device name is %s", devname)
    logger.info("Device version is %s", long_version)
    
    if is_mounted(udid):
        logger.info("<O_O> Developer Image has already mounted.")
        return

    image_file = download('DeveloperDiskImage.dmg', tmpdir, version)
    image_signature_file = download('DeveloperDiskImage.dmg.signature', tmpdir, version)

    mount_image(udid, image_file, image_signature_file)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import unicode_literals

import inspect
import logging
import os
import sys
import time
import threading
import datetime

from atx import strutils


class Logger(object):
    __alias = {
        'WARNING': 'WARN',
        'CRITICAL': 'FATAL'
    }

    def __init__(self, name=None, level=logging.INFO):
        if name is None:
            name = '-'
        self._name = name
        self._level = level
        self._lock = threading.Lock()

    def _write(self, s):
        self._lock.acquire()
        sys.stdout.write(s.rstrip() + '\n')
        self._lock.release()

    def setLevel(self, level):
        '''
        set format level

        Args:
            - level: for example, logging.INFO

        '''
        self._level = level
        return self

    def _level_write(self, level, str_format, *args):
        if level < self._level:
            return

        levelname = logging.getLevelName(level)
        message = str_format % args if args else str_format
        message = strutils.decode(message)
        frame, filename, line_number, function_name, lines, index = inspect.stack()[2]
        props = dict(
            asctime=datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f')[:-3],
            name=self._name,
            filename=os.path.basename(filename),
            lineno=line_number,
            message=message,
        )
        props['levelname'] = Logger.__alias.get(levelname, levelname)
        output = u'{asctime} {levelname:<5s} [{name}:{lineno:>4}] {message}'.format(**props)
        self._write(output)

    def debug(self, *args, **kwargs):
        self._level_write(logging.DEBUG, *args, **kwargs)

    def info(self, *args, **kwargs):
        self._level_write(logging.INFO, *args, **kwargs)

    def warn(self, *args, **kwargs):
        self._level_write(logging.WARN, *args, **kwargs)

    def error(self, *args, **kwargs):
        self._level_write(logging.ERROR, *args, **kwargs)

    def fatal(self, *args, **kwargs):
        self._level_write(logging.FATAL, *args, **kwargs)
        raise SystemExit(1)


def getLogger(name, level=logging.INFO):
    # logger = logging.getLogger(name)
    # ch = logging.StreamHandler()
    # fmt = "%(asctime)s %(levelname)-8.8s [%(name)s:%(lineno)4s] %(message)s"
    # ch.setFormatter(logging.Formatter(fmt))
    # ch.setLevel(level)
    # logger.handlers = [ch]
    return Logger(name, level=level)


if __name__ == '__main__':
    log = getLogger('test')
    log.debug("Should not see it.")
    log.setLevel(logging.DEBUG)
    log.setLevel(logging.DEBUG)
    log.info("This is info message")
    log.debug("This is debug message")
    log = getLogger('test')
    log.warn("This is warning message")
    log.error("This is error message")
    log.fatal("This is fatal message")
